/*
 * Copyright (c) 1991, 1994, 1995 Berkeley Software Design, Inc.
 * All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI kern_acct.c,v 2.11 1996/08/13 18:44:24 karels Exp
 */

/*-
 * Copyright (c) 1982, 1986, 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 * (c) UNIX System Laboratories, Inc.
 * All or some portions of this file are derived from material licensed
 * to the University of California by American Telephone and Telegraph
 * Co. or Unix System Laboratories, Inc. and are reproduced herein with
 * the permission of UNIX System Laboratories, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	from: @(#)kern_acct.c 8.8 (Berkeley) 5/14/95
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/namei.h>
#include <sys/resourcevar.h>
#include <sys/proc.h>
#include <sys/mount.h>
#include <sys/vnode.h>
#include <sys/file.h>
#include <sys/acct.h>
#include <sys/tty.h>
#include <sys/syslog.h>
#include <sys/kernel.h>
#include <sys/semaphore.h>

#include "syscallargs.h"

/*
 * SHOULD REPLACE THIS WITH A DRIVER THAT CAN BE READ TO SIMPLIFY.
 */
struct	vnode *acctp;		/* file of accounting records */
struct	ucred *acctcred;	/* credentials for writing acctp */

/*
 * Values associated with enabling and disabling accounting
 */
int	acctsuspend = 2;	/* stop accounting when < 2% free space left */
int	acctresume = 4;		/* resume when free space risen to > 4% */
int	acctchkfreq = 15;	/* frequency (in seconds) to check space */

static int accounting_disabled;
static void check_acct_fs_space __P((void));
static comp_t tv_to_comp __P((struct timeval *));

/*
 * Enable or disable process accounting.
 *
 * If a non-null filename is given, that file is used to store accounting
 * records on process exit. If a null filename is given process accounting
 * is terminated. If accounting is enabled, the system checks the amount
 * of freespace on the filesystem at timeval intervals. If the amount of
 * freespace is below acctsuspend percent, accounting is suspended. If
 * accounting has been suspended, and freespace rises above acctresume,
 * accounting is resumed.
 */
int
acct(p, uap, retval)
	struct proc *p;
	struct acct_args /* {
		syscallarg(char *) path;
	} */ *uap;
	int *retval;
{
	struct nameidata nd;
	struct vnode *vp = NULL;
	struct ucred *cr = NULL;
	struct vnode *oacctp;
	struct ucred *oacctcred;
	int error;

	if ((error = suser(p->p_cred->pc_ucred, NULL)) != 0)
		return (error);

	if (SCARG(uap, path) != NULL) {
		/*
		 * Allocate a vnode by 'opening' the file.
		 */
		NDINIT(&nd, LOOKUP, FOLLOW, UIO_USERSPACE, SCARG(uap, path), p);
		if ((error = vn_open(&nd, FFLAGS(O_WRONLY), 0)) != 0)
			return (error);
		vp = nd.ni_vp;
		VOP_UNLOCK(vp, 0, p);
		cr = p->p_ucred;
		crhold(cr);
	}

	/*
	 * Atomically swap accounting files.  This guarantees that we don't
	 * get into a race with acct_process(), and that we don't lose any
	 * accounting.  This could be simplified, but it's clearer this way.
	 */
	oacctp = acctp;
	oacctcred = acctcred;
	acctp = vp;
	acctcred = cr;

	if (oacctp != NULL) {
		(void)vn_close(oacctp, FWRITE, oacctcred, p);
		crfree(oacctcred);
	}

	check_acct_fs_space();

	return (0);
}

/*
 * Write an accounting record to the accounting file.  This code doesn't need
 * a lock since crucial structures are on the stack.
 */
int
acct_process(p)
	struct proc *p;
{
	struct acct a;
	struct timeval tv, utime, stime;
	struct pstats *psp = p->p_stats;
	struct rusage *rup;
	struct vnode *vp;
	u_int i;
	int error;

	check_acct_fs_space();

	if (acctp == NULL || accounting_disabled)
		return (0);

	strncpy(a.ac_comm, p->p_comm, sizeof(a.ac_comm));
	calcru(p, &utime, &stime, NULL);
	a.ac_utime = tv_to_comp(&utime);
	a.ac_stime = tv_to_comp(&stime);
	do {
		tv = time;
	} while (tv.tv_sec != time.tv_sec);
	timevalsub(&tv, &psp->p_start);
	a.ac_etime = tv_to_comp(&tv);
	a.ac_btime = psp->p_start.tv_sec;
	a.ac_uid = p->p_cred->p_ruid;
	a.ac_gid = p->p_cred->p_rgid;
	rup = &psp->p_ru;
	/* XXX should we bother to preserve low order time bits? */
	if ((i = (tv.tv_sec * AHZ) + (tv.tv_usec * AHZ) / 1000000) != 0)
		a.ac_mem =
		    ((rup->ru_ixrss + rup->ru_idrss + rup->ru_isrss) * AHZ) / i;
	else
		a.ac_mem = 0;
	/* The magic 64 comes from section 3.9 of the BSD book. */
	tv.tv_sec = (rup->ru_inblock + rup->ru_oublock) * 64;
	tv.tv_usec = 0;
	a.ac_io = tv_to_comp(&tv);
	if (p->p_pgrp->pg_session->s_ttyp)
		a.ac_tty = p->p_pgrp->pg_session->s_ttyp->t_dev;
	else
		a.ac_tty = NODEV;
	a.ac_flag = p->p_acflag;

	vp = acctp;
	VOP_LEASE(vp, p, acctcred, LEASE_WRITE);

	/* If the write fails, we assume we're out of space. */
	error = vn_rdwr(UIO_WRITE, vp, (caddr_t)&a, sizeof(struct acct),
	   (off_t)0, UIO_SYSSPACE, IO_UNIT|IO_APPEND, acctcred, NULL, NULL);
	if (error)
		accounting_disabled = 1;
	return (error);
}

/*
 * The first macro tests true if we have run out of space on the accounting
 * file's filesystem; the second tests true if we can turn it back on.
 */
#define	ACCT_FS_INSUFFICIENT_SPACE(sfp) \
	((sfp)->f_bavail <= acctsuspend * (sfp)->f_blocks / 100)
#define	ACCT_FS_ENOUGH_SPACE(sfp) \
	((sfp)->f_bavail > acctresume * (sfp)->f_blocks / 100)

/*
 * Check the space available to the accounting file.  Disable accounting if
 * there's too little, re-enable if it comes back.  Also disable accounting
 * if the vnode has been revoked or we can't do a statfs on its filesystem.
 * Beware the case where the vnode has been vgone()'d out from underneath us,
 * e.g. when the file system containing the accounting file has been forcibly
 * unmounted.
 */
static void
check_acct_fs_space()
{
	struct statfs sb;
	static long nextcheck;

	/* If not doing acounting, or we checked recently, just return. */
	if (acctp == NULL || time.tv_sec < nextcheck)
		return;
	nextcheck = time.tv_sec + acctchkfreq;

	if (acctp->v_type == VBAD) {
		(void)vn_close(acctp, FWRITE, NOCRED, NULL);
		acctp = NULL;
		return;
	}
	if (VFS_STATFS(acctp->v_mount, &sb, (struct proc *)0)) {
		accounting_disabled = 1;
		return;
	}
	if (accounting_disabled) {
		if (ACCT_FS_ENOUGH_SPACE(&sb)) {
			accounting_disabled = 0;
			log(LOG_NOTICE, "Accounting resumed\n");
		}
	} else {
		if (ACCT_FS_INSUFFICIENT_SPACE(&sb)) {
			accounting_disabled = 1;
			log(LOG_NOTICE, "Accounting suspended\n");
		}
	}
}

/*
 * ASSUMES TWO'S COMPLEMENT ARITHMETIC.
 */
static comp_t
tv_to_comp(tvp)
	struct timeval *tvp;
{
	register int exp;
	u_long mant;
#define MANTBITS 13
#define EXPBITS 3

	/*
	 * Raise exponent and drop bits off the right of the mantissa
	 * until the mantissa fits in its field.  If we run out of exponent
	 * space, return the maximum possible value.  With an AHZ value
	 * of 64, this is good for up to 8.5 years.
	 * (8191 * (1 << (3*7)) / 64 / 60 / 60 / 24 / 365.242 ~= 8.5)
	 */
	mant = tvp->tv_sec * AHZ + (tvp->tv_usec * AHZ) / 1000000;
	for (exp = 0; exp < (1 << EXPBITS); exp++, mant >>= EXPBITS)
		if (mant < (1 << MANTBITS))
			return (mant | (exp << MANTBITS));
	return (~0);
}
