/*-
 * Copyright (c) 1991, 1994, 1995, 1996, 1997 Berkeley Software Design, Inc.
 * All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI kern_physio.c,v 2.6 1997/01/16 05:24:17 karels Exp
 */

/*
 * Copyright (c) 1982, 1986, 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	from: @(#)kern_physio.c	7.20 (Berkeley) 5/11/91
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/proc.h>
#include <sys/trace.h>
#include <sys/map.h>
#include <sys/vnode.h>
#include <sys/malloc.h>

#include <vm/vm.h>
#include <vm/vm_kern.h>

int
physio_start_vm(map, uaddr, bp, mlistp)
	vm_map_t map;
	vm_offset_t uaddr;
	struct buf *bp;
	vm_page_t **mlistp;
{
	vm_offset_t start = trunc_page(uaddr);
	vm_offset_t v, p;
	vm_page_t *mp, *mp2, *mlist;
	vm_size_t len = round_page(uaddr + bp->b_bcount) - start;
	int error = 0;

	if (len == 0)
		panic("physio_start_vm zero transfer");

	/* Lock the pages in memory.  */
	if ((error = vm_map_pageable(map, start, start + len, FALSE)) != 0)
		return (error);

	/*
	 * Create a writable kernel alias for the pages.
	 * We use the tools that pagers use for mapping pages.
	 * We mark the pages busy so that no other process
	 * can perform I/O on them while we fill or drain them.
	 */
	MALLOC(mlist, vm_page_t *, atop(len) * sizeof (mlist[0]),
	    M_TEMP, M_WAITOK);
again:
	for (mp = &mlist[0], v = start; v < start + len; v += PAGE_SIZE, ++mp) {
		p = pmap_extract(map->pmap, v);
		*mp = PHYS_TO_VM_PAGE(p);
		if ((*mp)->flags & PG_BUSY) {

			/* Did we grab this page already?  */
			for (mp2 = &mlist[0]; *mp2 != *mp; ++mp2)
				;
			if (mp2 != mp)
				continue;

			/*
			 * If we can't acquire all the pages,
			 * bag it and try again later.
			 */
			for (mp2 = &mlist[0]; mp2 < mp; ++mp2) {
				(*mp2)->flags &= ~PG_BUSY;
				PAGE_WAKEUP(*mp);
			}
			PAGE_ASSERT_WAIT(*mp, TRUE);
			thread_block();
			goto again;
		}
		(*mp)->flags |= PG_BUSY;

		/*
		 * Keep other processes from writing into the pages
		 * before the transfer can be made.
		 * It'd be nice if we could prevent reading too,
		 * but then we would be unable to use vm_map_pageable()
		 * to restore the region when we finished.
		 */
		pmap_page_protect(p, VM_PROT_READ);
	}

	v = vm_pager_map_pages(mlist, atop(len), 1);

	bp->b_un.b_addr = (caddr_t)(v + uaddr - start);

	*mlistp = mlist;

	return (0);
}

/*
 * This takes a buf pointer AND a count because we can't trust
 * the buffer's count field after an I/O.
 */
void
physio_finish_vm(map, uaddr, bp, cnt, mlist)
	vm_map_t map;
	vm_offset_t uaddr;
	struct buf *bp;
	vm_size_t cnt;
	vm_page_t *mlist;
{
	vm_offset_t kpage = trunc_page(bp->b_un.b_addr);
	vm_offset_t start = trunc_page(uaddr);
	vm_size_t len = round_page(uaddr + cnt) - start;
	int i;

	vm_pager_unmap_pages(kpage, atop(len));

	for (i = 0; i < atop(len); ++i) {
		mlist[i]->flags &= ~PG_BUSY;
		PAGE_WAKEUP(mlist[i]);
	}
	FREE(mlist, M_TEMP);

	vm_map_pageable(map, start, start + len, TRUE);
}

int
physio(devio, bufhdr, deviceid, ioflag, sizelimit, uio)
	int (*devio)(); 
	register struct buf *bufhdr;
	dev_t deviceid;
	int ioflag;
	void (*sizelimit) __P((struct buf *));
	struct uio *uio;
{
	vm_map_t map = &uio->uio_procp->p_vmspace->vm_map;
	vm_page_t *mlist;
	vm_offset_t base;
	vm_size_t cnt, req;
	struct iovec *iov;
	int s;
	int error = 0;

	/*
	 * Complete pseudo-code for physio() can be found on
	 * p. 232 of the BSD book (figure 8.1).  Oh joy!
	 */

	if (uio->uio_segflg != UIO_USERSPACE)
		panic("physio uio_segflg");

	/*
	 * 4.3 Tahoe evidently 'borrow[ed] swap buffer structures' (p. 232);
	 * we just malloc() bufs when they aren't provided.
	 */
	if (bufhdr) {
		s = splbio();
		while (bufhdr->b_flags & B_BUSY) {
			bufhdr->b_flags |= B_WANTED;
			sleep((caddr_t)bufhdr, PRIBIO + 1);
		}
		bufhdr->b_flags |= B_BUSY | B_PHYS | ioflag;
		bufhdr->b_flags &= ~(B_WANTED | B_DONE);
		splx(s);
	} else {
		MALLOC(bufhdr, struct buf *, sizeof *bufhdr, M_BUFFER,
		    M_WAITOK);
		bufhdr->b_flags = B_BUSY | ioflag;
	}

	if ((uio->uio_rw == UIO_READ) == ((bufhdr->b_flags & B_READ) == 0))
		panic("physio inconsistent directions");

	bufhdr->b_error = 0;
	bufhdr->b_dev = deviceid;
	bufhdr->b_resid = 0;
	bufhdr->b_chain = NULL;
	bufhdr->b_proc = uio->uio_procp;
	bufhdr->b_iodone = NULL;
	bufhdr->b_vp = NULL;

	/*
	 * The uio/iov idiom is modeled on uiomove() in kern_subr.c.
	 */
	while (uio->uio_resid) {
		iov = uio->uio_iov;
		base = (vm_offset_t) iov->iov_base;
		cnt = iov->iov_len;
		if (cnt == 0) {
			uio->uio_iov++;
			uio->uio_iovcnt--;
			continue;
		}

		/*
		 * Adjust the size of the transfer.
		 */
		bufhdr->b_bcount = cnt;
		(*sizelimit)(bufhdr);
		req = cnt = bufhdr->b_bcount;

		/*
		 * Check for attempts to penetrate kernel space.
		 * ??? WHY ARE WE DOING THIS RATHER THAN USING useracc?
		 */
		if (base + cnt > VM_MAXUSER_ADDRESS)
			error = EFAULT;
		if (VM_MIN_ADDRESS > 0 && base < VM_MIN_ADDRESS)
			error = EFAULT;
		if (base > base + cnt)
			error = EFAULT;
		if (error)
			break;

		if ((error = physio_start_vm(map, base, bufhdr, &mlist)) != 0)
			break;

		/*
		 * Perform the transfer.
		 */
		bufhdr->b_blkno = btodb(uio->uio_offset);
		bufhdr->b_iocount = cnt;
		(*devio)(bufhdr);
		error = biowait(bufhdr);
		bufhdr->b_flags &= ~B_DONE;
		physio_finish_vm(map, base, bufhdr, cnt, mlist);

		/*
		 * The driver might decrease b_bcount and/or set b_resid.
		 */
		cnt = bufhdr->b_bcount - bufhdr->b_resid;

		iov->iov_base += cnt;
		iov->iov_len -= cnt;
		uio->uio_resid -= cnt;
		uio->uio_offset += cnt;

		if (error)
			break;

		if (cnt != req)
			/* EOF or a short read */
			break;
	}

	if (bufhdr->b_flags & B_PHYS) {
		s = splbio();
		bufhdr->b_flags &= ~B_BUSY;
		if (bufhdr->b_flags & B_WANTED) {
			bufhdr->b_flags &= ~B_WANTED;
			wakeup((caddr_t) bufhdr);
		}
		splx(s);
	} else
		FREE(bufhdr, M_BUFFER);

	return (error);
}

/*
 * Calculate the maximum size of I/O request that can be requested
 * in a single operation. This limit is necessary to prevent a single
 * process from being able to lock more than a fixed amount of memory
 * in the kernel.
 */
void
minphys(bufhdr)
	struct buf *bufhdr;
{

	bufhdr->b_bcount = min(bufhdr->b_bcount, MAXPHYS);
}

/*
 * Do a read on a device for a user process.
 */
int
rawread(deviceid, uio, flag)
	dev_t deviceid;
	struct uio *uio;
	int flag;
{
	return (physio(devsw[major(deviceid)]->d_strategy, (struct buf *)NULL,
	    deviceid, B_READ, minphys, uio));
}

/*
 * Do a write on a device for a user process.
 */
int
rawwrite(deviceid, uio, flag)
	dev_t deviceid;
	struct uio *uio;
	int flag;
{
	return (physio(devsw[major(deviceid)]->d_strategy, (struct buf *)NULL,
	    deviceid, B_WRITE, minphys, uio));
}
