/*-
 * Copyright (c) 1995, 1996 Berkeley Software Design, Inc. All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI kern_sysctl.c,v 2.22 1996/08/27 17:12:47 prb Exp
 */


/*-
 * Copyright (c) 1982, 1986, 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Karels at Berkeley Software Design, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)kern_sysctl.c	8.9 (Berkeley) 5/20/95
 */

/*
 * sysctl system call.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/proc.h>
#include <sys/file.h>
#include <sys/vnode.h>
#include <sys/namei.h>
#include <sys/unistd.h>
#include <sys/buf.h>
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <vm/vm.h>
#include <sys/sysctl.h>
#include <sys/cpustats.h>
#include <sys/disk.h>
#include <sys/diskstats.h>
#include <sys/ttystats.h>
#include <sys/ipc.h>
#include <sys/license.h>

#include <sys/mount.h>
#include <sys/semaphore.h>

#include "syscallargs.h"

sysctlfn kern_sysctl;
sysctlfn hw_sysctl;
#ifdef DEBUG
sysctlfn debug_sysctl;
#endif
extern sysctlfn vm_sysctl;
extern sysctlfn vfs_sysctl;
extern sysctlfn net_sysctl;
extern sysctlfn cpu_sysctl;
extern sysctlfn ipc_sysctl;

/*
 * Locking and stats
 */
static struct sysctl_lock {
	int	sl_lock;
	int	sl_want;
	int	sl_locked;
} memlock;

int
__sysctl(p, uap, retval)
	struct proc *p;
	register struct __sysctl_args /* {
		syscallarg(int *) name;
		syscallarg(u_int) namelen;
		syscallarg(void *) old;
		syscallarg(size_t *) oldlenp;
		syscallarg(void *) new;
		syscallarg(size_t) newlen;
	} */ *uap;
	register_t *retval;
{
	int error, dolock = 1;
	size_t savelen, oldlen = 0;
	sysctlfn *fn;
	int name[CTL_MAXNAME];

	if (SCARG(uap, new) != NULL &&
	    (error = suser(p->p_ucred, &p->p_acflag)))
		return (error);
	/*
	 * all top-level sysctl names are non-terminal
	 */
	if (SCARG(uap, namelen) > CTL_MAXNAME || SCARG(uap, namelen) < 2)
		return (EINVAL);
	if ((error = copyin(SCARG(uap, name),
	    &name, SCARG(uap, namelen) * sizeof(int))) != 0)
		return (error);

	switch (name[0]) {
	case CTL_KERN:
		fn = kern_sysctl;
		if (name[2] == KERN_VNODE)	/* XXX */
			dolock = 0;
		break;
	case CTL_HW:
		fn = hw_sysctl;
		break;
	case CTL_VM:
		fn = vm_sysctl;
		break;
	case CTL_NET:
		fn = net_sysctl;
		break;
	case CTL_VFS:
		fn = vfs_sysctl;
		break;
	case CTL_MACHDEP:
		fn = cpu_sysctl;
		break;
#ifdef DEBUG
	case CTL_DEBUG:
		fn = debug_sysctl;
		break;
#endif
	default:
		return (EOPNOTSUPP);
	}

	if (SCARG(uap, oldlenp) &&
	    (error = copyin(SCARG(uap, oldlenp), &oldlen, sizeof(oldlen))))
		return (error);
	if (SCARG(uap, old) != NULL) {
		if (!useracc(SCARG(uap, old), oldlen, B_WRITE))
			return (EFAULT);
		while (memlock.sl_lock) {
			memlock.sl_want = 1;
			sleep((caddr_t)&memlock, PRIBIO+1);
			memlock.sl_locked++;
		}
		memlock.sl_lock = 1;
		if (dolock)
			vslock(SCARG(uap, old), oldlen);
		savelen = oldlen;
	}
	error = (*fn)(name + 1, SCARG(uap, namelen) - 1, SCARG(uap, old),
	    &oldlen, SCARG(uap, new), SCARG(uap, newlen), p);
	if (SCARG(uap, old) != NULL) {
		if (dolock)
			vsunlock(SCARG(uap, old), savelen, B_WRITE);
		memlock.sl_lock = 0;
		if (memlock.sl_want) {
			memlock.sl_want = 0;
			wakeup((caddr_t)&memlock);
		}
	}
	if (error)
		return (error);
	if (SCARG(uap, oldlenp))
		error = copyout(&oldlen, SCARG(uap, oldlenp), sizeof(oldlen));

	/* Previous versions returned oldlen via retval, which is incorrect. */
	return (0);
}

/*
 * Attributes stored in the kernel.
 */
static int nisdomainnamelen;
static char nisdomainname[MAXHOSTNAMELEN];	/* NIS domain name. */
int hostnamelen;
char hostname[MAXHOSTNAMELEN];			/* Host name. */
long hostid;
int securelevel;

/*
 * kernel related system variables.
 */
int
kern_sysctl(name, namelen, oldp, oldlenp, newp, newlen, p)
	int *name;
	u_int namelen;
	void *oldp;
	size_t *oldlenp;
	void *newp;
	size_t newlen;
	struct proc *p;
{
	int error, level, inthostid;
	license_t lp;
	extern char ostype[], osrelease[], version[];
	extern struct ttytotals ttytotals;
	extern struct nchstats nchstats;
	extern int deadman, maxsymlinks;

	/* all sysctl names at this level are terminal */
	if (namelen != 1 && !(name[0] == KERN_PROC || name[0] == KERN_PROF || 
	    name[0] == KERN_SYSVIPC))
		return (ENOTDIR);		/* overloaded */

	switch (name[0]) {
	case KERN_OSTYPE:
		return (sysctl_rdstring(oldp, oldlenp, newp, ostype));
	case KERN_OSRELEASE:
		return (sysctl_rdstring(oldp, oldlenp, newp, osrelease));
	case KERN_OSREV:
		return (sysctl_rdint(oldp, oldlenp, newp, _BSDI_VERSION));
	case KERN_VERSION:
		return (sysctl_rdstring(oldp, oldlenp, newp, version));
	case KERN_MAXVNODES:
		return(sysctl_int(oldp, oldlenp, newp, newlen, &desiredvnodes));
	case KERN_MAXPROC:
		return (sysctl_int(oldp, oldlenp, newp, newlen, &maxproc));
	case KERN_MAXFILES:
		return (sysctl_int(oldp, oldlenp, newp, newlen, &maxfiles));
	case KERN_ARGMAX:
		return (sysctl_rdint(oldp, oldlenp, newp, ARG_MAX));
	case KERN_SECURELVL:
		level = securelevel;
		if ((error = sysctl_int(oldp, oldlenp, newp, newlen, &level)) ||
		    newp == NULL)
			return (error);
		if (level < securelevel && p->p_pid != 1)
			return (EPERM);
		securelevel = level;
		return (0);
	case KERN_NISDOMAINNAME:
		error = sysctl_string(oldp, oldlenp, newp, newlen,
		    nisdomainname, sizeof(nisdomainname));
		if (newp && !error)
			nisdomainnamelen = newlen;
		return (error);
	case KERN_HOSTNAME:
		error = sysctl_string(oldp, oldlenp, newp, newlen,
		    hostname, sizeof(hostname));
		if (newp && !error)
			hostnamelen = newlen;
		return (error);
	case KERN_HOSTID:
		inthostid = hostid;  /* XXX assumes sizeof long <= sizeof int */
		error =  sysctl_int(oldp, oldlenp, newp, newlen, &inthostid);
		hostid = inthostid;
		return (error);
	case KERN_CLOCKRATE:
		return (sysctl_clockrate(oldp, oldlenp));
	case KERN_BOOTTIME:
		return (sysctl_rdstruct(oldp, oldlenp, newp, &boottime,
		    sizeof(struct timeval)));
	case KERN_VNODE:
		return (sysctl_vnode(oldp, oldlenp, p));
	case KERN_PROC:
		return (sysctl_doproc(name + 1, namelen - 1, oldp, oldlenp));
	case KERN_FILE_COMPAT:				/* XXX compat */
		return (sysctl_file(oldp, oldlenp, 1));	/* XXX compat */
	case KERN_FILES:
		return (sysctl_file(oldp, oldlenp, 0));
#ifdef GPROF
	case KERN_PROF:
		return (sysctl_doprof(name + 1, namelen - 1, oldp, oldlenp,
		    newp, newlen));
#endif
	case KERN_POSIX1:
		return (sysctl_rdint(oldp, oldlenp, newp, _POSIX_VERSION));
	case KERN_NGROUPS:
		return (sysctl_rdint(oldp, oldlenp, newp, NGROUPS_MAX));
	case KERN_JOB_CONTROL:
		return (sysctl_rdint(oldp, oldlenp, newp, 1));
	case KERN_SAVED_IDS:
		return (sysctl_rdint(oldp, oldlenp, newp, 0));
	case KERN_CPUSTATS:
		return (sysctl_rdstruct(oldp, oldlenp, newp,
		    &cpustats, sizeof cpustats));
	case KERN_TTYTOTALS:
		return (sysctl_rdstruct(oldp, oldlenp, newp,
		    &ttytotals, sizeof ttytotals));
	case KERN_NCHSTATS:
		return (sysctl_rdstruct(oldp, oldlenp, newp,
		    &nchstats, sizeof nchstats));
	case KERN_KMEMBUCKETS:
		return (sysctl_rdstruct(oldp, oldlenp, newp,
		    bucket, sizeof bucket[0] * (MINBUCKET + 16)));
	case KERN_KMEMSTATS:
		return (sysctl_rdstruct(oldp, oldlenp, newp,
		    kmemstats, sizeof kmemstats[0] * M_LAST));
	case KERN_DEADMANTIMER:
		return (sysctl_int(oldp, oldlenp, newp, newlen, &deadman));
	case KERN_MAXSYMLINKS:
		return (sysctl_int(oldp, oldlenp, newp, newlen, &maxsymlinks));
	case KERN_SYSVIPC:
		return (ipc_sysctl(name + 1, namelen - 1, oldp, oldlenp,
		    newp, newlen, p));
	case KERN_LICENSE_USERS:
		return (sysctl_rdint(oldp, oldlenp, newp, license_users()));
	case KERN_LICENSE_FLAGS:
		return (sysctl_rdint(oldp, oldlenp, newp, license_flags()));
	case KERN_LICENSE:
		if (oldp) {
			if (error = suser(p->p_ucred, &p->p_acflag))
				return (error);
			lp = *license_id();
		}
		if (error = sysctl_struct(oldp, oldlenp, newp, newlen,
		    &lp, sizeof(lp)))
			return (error);
		if (newp)
			return (license_set(&lp));
		return (0);
	default:
		return (EOPNOTSUPP);
	}
	/* NOTREACHED */
}

/*
 * hardware related system variables.
 */
int
hw_sysctl(name, namelen, oldp, oldlenp, newp, newlen, p)
	int *name;
	u_int namelen;
	void *oldp;
	size_t *oldlenp;
	void *newp;
	size_t newlen;
	struct proc *p;
{
	extern char machine[], cpu_model[];

	/* all sysctl names at this level are terminal */
	if (namelen != 1)
		return (ENOTDIR);		/* overloaded */

	switch (name[0]) {
	case HW_MACHINE:
		return (sysctl_rdstring(oldp, oldlenp, newp, machine));
	case HW_MODEL:
		return (sysctl_rdstring(oldp, oldlenp, newp, cpu_model));
	case HW_NCPU:
		return (sysctl_rdint(oldp, oldlenp, newp, 1));	/* XXX */
	case HW_BYTEORDER:
		return (sysctl_rdint(oldp, oldlenp, newp, BYTE_ORDER));
	case HW_PHYSMEM:
		return (sysctl_rdint(oldp, oldlenp, newp, ctob(physmem)));
	case HW_USERMEM:
		return (sysctl_rdint(oldp, oldlenp, newp,
		    ctob(physmem - cnt.v_wire_count)));
	case HW_PAGESIZE:
		return (sysctl_rdint(oldp, oldlenp, newp, PAGE_SIZE));
	case HW_DISKNAMES:
		return (sysctl_disknames(oldp, oldlenp));
	case HW_DISKSTATS:
		return (sysctl_diskstats(oldp, oldlenp));
	case HW_TTYNAMES_TMP:
		return (sysctl_ttynames(oldp, oldlenp));
	case HW_TTYSTATS_TMP:
		return (sysctl_ttystats(oldp, oldlenp));
	default:
		return (EOPNOTSUPP);
	}
	/* NOTREACHED */
}

#ifdef DEBUG
/*
 * Debugging related system variables.
 */
struct ctldebug debug0, debug1, debug2, debug3, debug4;
struct ctldebug debug5, debug6, debug7, debug8, debug9;
struct ctldebug debug10, debug11, debug12, debug13, debug14;
struct ctldebug debug15, debug16, debug17, debug18, debug19;
static struct ctldebug *debugvars[CTL_DEBUG_MAXID] = {
	&debug0, &debug1, &debug2, &debug3, &debug4,
	&debug5, &debug6, &debug7, &debug8, &debug9,
	&debug10, &debug11, &debug12, &debug13, &debug14,
	&debug15, &debug16, &debug17, &debug18, &debug19,
};
int
debug_sysctl(name, namelen, oldp, oldlenp, newp, newlen, p)
	int *name;
	u_int namelen;
	void *oldp;
	size_t *oldlenp;
	void *newp;
	size_t newlen;
	struct proc *p;
{
	struct ctldebug *cdp;

	/* all sysctl names at this level are name and field */
	if (namelen != 2)
		return (ENOTDIR);		/* overloaded */
	cdp = debugvars[name[0]];
	if (name[0] >= CTL_DEBUG_MAXID || cdp->debugname == 0)
		return (EOPNOTSUPP);
	switch (name[1]) {
	case CTL_DEBUG_NAME:
		return (sysctl_rdstring(oldp, oldlenp, newp, cdp->debugname));
	case CTL_DEBUG_VALUE:
		return (sysctl_int(oldp, oldlenp, newp, newlen, cdp->debugvar));
	default:
		return (EOPNOTSUPP);
	}
	/* NOTREACHED */
}
#endif /* DEBUG */

/*
 * Validate parameters and get old / set new parameters
 * for an integer-valued sysctl function.
 */
int
sysctl_int(oldp, oldlenp, newp, newlen, valp)
	void *oldp;
	size_t *oldlenp;
	void *newp;
	size_t newlen;
	int *valp;
{
	int error = 0;

	if (oldp && *oldlenp < sizeof(int))
		return (ENOMEM);
	if (newp && newlen != sizeof(int))
		return (EINVAL);
	*oldlenp = sizeof(int);
	if (oldp)
		error = copyout(valp, oldp, sizeof(int));
	if (error == 0 && newp)
		error = copyin(newp, valp, sizeof(int));
	return (error);
}

/*
 * As above, but read-only.
 */
int
sysctl_rdint(oldp, oldlenp, newp, val)
	void *oldp;
	size_t *oldlenp;
	void *newp;
	int val;
{
	int error = 0;

	if (oldp && *oldlenp < sizeof(int))
		return (ENOMEM);
	if (newp)
		return (EPERM);
	*oldlenp = sizeof(int);
	if (oldp)
		error = copyout((caddr_t)&val, oldp, sizeof(int));
	return (error);
}

int
sysctl_int_arr(valp, name, namelen, oldp, oldlenp, newp, newlen)
	int **valp;
	int *name;
	u_int namelen;
	void *oldp;
	size_t *oldlenp;
	void *newp;
	size_t newlen;
{

	/* All sysctl names at this level are terminal. */
	if (namelen != 1)
		return (ENOTDIR);

	valp += name[0];
	if (*valp)
		return (sysctl_int(oldp, oldlenp, newp, newlen, *valp));
	else
		return (EOPNOTSUPP);
}

/*
 * Validate parameters and get old / set new parameters
 * for a string-valued sysctl function.
 */
int
sysctl_string(oldp, oldlenp, newp, newlen, str, maxlen)
	void *oldp;
	size_t *oldlenp;
	void *newp;
	size_t newlen;
	char *str;
	int maxlen;
{
	int len, error = 0;
	int small_buf = 0;

	len = strlen(str) + 1;
	if (oldp && *oldlenp < len) {
		small_buf = 1;
		len = *oldlenp;
	}
	if (newp && newlen >= maxlen)
		return (EINVAL);
	if (oldp) {
		*oldlenp = len;
		error = copyout(str, oldp, len);
	}
	if (error == 0 && newp) {
		error = copyin(newp, str, newlen);
		if (error == 0)
			str[newlen] = 0;
	}
	if (error == 0 && small_buf)
		return (ENOMEM);
	return (error);
}

/*
 * As above, but read-only.
 */
int
sysctl_rdstring(oldp, oldlenp, newp, str)
	void *oldp;
	size_t *oldlenp;
	void *newp;
	char *str;
{
	int len, error = 0;
	int small_buf = 0;

	len = strlen(str) + 1;
	if (oldp && *oldlenp < len) {
		small_buf = 1;
		len = *oldlenp;
	}
	if (newp)
		return (EPERM);
	*oldlenp = len;
	if (oldp)
		error = copyout(str, oldp, len);
	if (error == 0 && small_buf)
		return (ENOMEM);
	return (error);
}

/*
 * Validate parameters and get old / set new parameters
 * for a structure oriented sysctl function.
 */
int
sysctl_struct(oldp, oldlenp, newp, newlen, sp, len)
	void *oldp;
	size_t *oldlenp;
	void *newp;
	size_t newlen;
	void *sp;
	int len;
{
	int error = 0;

	if (oldp && *oldlenp < len)
		return (ENOMEM);
	if (newp && newlen > len)
		return (EINVAL);
	if (oldp) {
		*oldlenp = len;
		error = copyout(sp, oldp, len);
	}
	if (error == 0 && newp)
		error = copyin(newp, sp, len);
	return (error);
}

/*
 * Validate parameters and get old parameters
 * for a structure oriented sysctl function.
 */
int
sysctl_rdstruct(oldp, oldlenp, newp, sp, len)
	void *oldp;
	size_t *oldlenp;
	void *newp, *sp;
	int len;
{
	int error = 0;

	if (oldp && *oldlenp < len)
		return (ENOMEM);
	if (newp)
		return (EPERM);
	*oldlenp = len;
	if (oldp)
		error = copyout(sp, oldp, len);
	return (error);
}

/*
 * Validate parameters and get old parameters
 * for a structure oriented sysctl function.
 */
sysctl_rdtrunc(oldp, oldlenp, newp, sp, len)
	void *oldp;
	size_t *oldlenp;
	void *newp, *sp;
	int len;
{
	int error = 0;

	if (oldp && *oldlenp < len)
		len = *oldlenp;
	if (newp)
		return (EPERM);
	*oldlenp = len;
	if (oldp)
		error = copyout(sp, oldp, len);
	return (error);
}

/*
 * Get file structures.
 */
int
sysctl_file(where, sizep, compat)
	char *where;
	size_t *sizep;
	int compat;
{
	int buflen, error;
	struct file *fp, ftemp;
	char *start = where;

	buflen = *sizep;
	if (where == NULL) {
		/*
		 * overestimate by 10 files
		 */
		*sizep = (compat ? sizeof(filehead) : 0) +
		    (nfiles + 10) * sizeof(struct file);
		return (0);
	}

	if (compat) {
		/* Put filehead at front. */
		if (buflen < sizeof(filehead)) {
			*sizep = 0;
			return (0);
		}
		if ((error = copyout(&filehead, where, sizeof(filehead))) != 0)
			return (error);
		buflen -= sizeof(filehead);
		where += sizeof(filehead);
	}

	/* Return an array of file structures. */
	for (fp = filehead.lh_first; fp != 0; fp = fp->f_list.le_next) {
		if (buflen < sizeof(struct file)) {
			*sizep = where - start;
			return (ENOMEM);
		}
		/* User code gets a self-pointer rather than a forward. */
		ftemp = *fp;
		if (!compat)
			ftemp.f_list.le_next = fp;
		if ((error = copyout(&ftemp, where, sizeof(ftemp))) != 0)
			return (error);
		buflen -= sizeof(ftemp);
		where += sizeof(ftemp);
	}
	*sizep = where - start;
	return (0);
}

/*
 * Return a list of disk names, each null terminated,
 * each starting after the previous null, in the same order
 * as the disk statistics will be returned.
 */
int
sysctl_disknames(where, sizep)
	char *where;
	size_t *sizep;
{
	register struct dkdevice *dk;
	size_t namelen, buflen, needed;
	char *start;
	int error;

	buflen = *sizep;
	needed = 0;
	start = where;
	for (dk = diskhead; dk != NULL; dk = dk->dk_next) {
		namelen = strlen(dk->dk_dev.dv_xname) + 1;
		if (where == NULL) {
			needed += namelen;
			continue;
		}
		if (buflen < namelen)
			break;
		if ((error = copyout(dk->dk_dev.dv_xname, where, namelen)) != 0)
			return (error);
		buflen -= namelen;
		where += namelen;
	}
	*sizep = where == NULL ? needed : where - start;
	return (0);
}

/*
 * Get disk statistics.
 */
int
sysctl_diskstats(where, sizep)
	char *where;
	size_t *sizep;
{
	register struct dkdevice *dk;
	size_t buflen, needed;
	char *start;
	int error;

	buflen = *sizep;
	needed = 0;
	start = where;
	for (dk = diskhead; dk != NULL; dk = dk->dk_next) {
		if (where == NULL) {
			needed += sizeof dk->dk_stats;
			continue;
		}
		if (buflen < sizeof dk->dk_stats)
			break;
		if ((error =
		    copyout(&dk->dk_stats, where, sizeof(dk->dk_stats))) != 0)
			return (error);
		buflen -= sizeof dk->dk_stats;
		where += sizeof dk->dk_stats;
	}
	*sizep = where == NULL ? needed : where - start;
	return (0);
}

/*
 * try over estimating by 5 procs
 */
#define KERN_PROCSLOP	(5 * sizeof (struct kinfo_proc))

int
sysctl_doproc(name, namelen, where, sizep)
	int *name;
	u_int namelen;
	char *where;
	size_t *sizep;
{
	register struct proc *p;
	register struct kinfo_proc *dp = (struct kinfo_proc *)where;
	register int needed = 0;
	int buflen = where != NULL ? *sizep : 0;
	int doingzomb;
	struct eproc eproc;
	int error = 0;

	if (namelen != 2 && !(namelen == 1 && name[0] == KERN_PROC_ALL))
		return (EINVAL);
	p = allproc.lh_first;
	doingzomb = 0;
again:
	for (; p != 0; p = p->p_list.le_next) {
		/*
		 * Skip embryonic processes.
		 */
		if (p->p_stat == SIDL)
			continue;
		/*
		 * TODO - make more efficient (see notes below).
		 * do by session.
		 */
		switch (name[0]) {

		case KERN_PROC_PID:
			/* could do this with just a lookup */
			if (p->p_pid != (pid_t)name[1])
				continue;
			break;

		case KERN_PROC_PGRP:
			/* could do this by traversing pgrp */
			if (p->p_pgrp->pg_id != (pid_t)name[1])
				continue;
			break;

		case KERN_PROC_TTY:
			if ((p->p_flag & P_CONTROLT) == 0 ||
			    p->p_session->s_ttyp == NULL ||
			    p->p_session->s_ttyp->t_dev != (dev_t)name[1])
				continue;
			break;

		case KERN_PROC_UID:
			if (p->p_ucred->cr_uid != (uid_t)name[1])
				continue;
			break;

		case KERN_PROC_RUID:
			if (p->p_cred->p_ruid != (uid_t)name[1])
				continue;
			break;
		}
		if (buflen >= sizeof(struct kinfo_proc)) {
			fill_eproc(p, &eproc);
			if ((error = copyout(p,
			    &dp->kp_proc, sizeof(struct proc))) != 0)
				return (error);
			if ((error = copyout(&eproc,
			    &dp->kp_eproc, sizeof(eproc))) != 0)
				return (error);
			dp++;
			buflen -= sizeof(struct kinfo_proc);
		}
		needed += sizeof(struct kinfo_proc);
	}
	if (doingzomb == 0) {
		p = zombproc.lh_first;
		doingzomb++;
		goto again;
	}
	if (where != NULL) {
		*sizep = (caddr_t)dp - where;
		if (needed > *sizep)
			return (ENOMEM);
	} else {
		needed += KERN_PROCSLOP;
		*sizep = needed;
	}
	return (0);
}

/*
 * Fill in an eproc structure for the specified process.
 */
void
fill_eproc(p, ep)
	register struct proc *p;
	register struct eproc *ep;
{
	register struct tty *tp;

	ep->e_paddr = p;
	ep->e_sess = p->p_pgrp->pg_session;
	ep->e_pcred = *p->p_cred;
	ep->e_ucred = *p->p_ucred;
	if (p->p_stat == SIDL || p->p_stat == SZOMB)
		bzero(&ep->e_vm, sizeof (ep->e_vm));
	else {
		register struct vmspace *vm = p->p_vmspace;

#ifdef sparc
		ep->e_vm.vm_tsize = vm->vm_tsize;
		ep->e_vm.vm_dsize = vm->vm_dsize;
		ep->e_vm.vm_ssize = vm->vm_ssize;
#else
		ep->e_vm = *p->p_vmspace;
#endif
#ifdef pmap_resident_count
		ep->e_vm.vm_rssize = pmap_resident_count(&vm->vm_pmap); /*XXX*/
#else
#ifdef sparc
		ep->e_vm.vm_rssize = vm->vm_rssize;
#endif
#endif
	}
	if (p->p_pptr)
		ep->e_ppid = p->p_pptr->p_pid;
	else
		ep->e_ppid = 0;
	ep->e_pgid = p->p_pgrp->pg_id;
	ep->e_jobc = p->p_pgrp->pg_jobc;
	if ((p->p_flag & P_CONTROLT) &&
	     (tp = ep->e_sess->s_ttyp)) {
		ep->e_tdev = tp->t_dev;
		ep->e_tpgid = tp->t_pgrp ? tp->t_pgrp->pg_id : NO_PID;
		ep->e_tsess = tp->t_session;
	} else
		ep->e_tdev = NODEV;
	ep->e_flag = ep->e_sess->s_ttyvp ? EPROC_CTTY : 0;
	if (SESS_LEADER(p))
		ep->e_flag |= EPROC_SLEADER;
	if (p->p_wmesg)
		strncpy(ep->e_wmesg, p->p_wmesg, WMESGLEN);
	strncpy(ep->e_login, p->p_session->s_login, sizeof(ep->e_login));
	ep->e_xsize = ep->e_xrssize = 0;
	ep->e_xccount = ep->e_xswrss = 0;
}

#ifdef COMPAT_GETKERNINFO
#include <sys/socket.h>
#include <sys/sysinfo.h>
#define	KINFO_PROC		(0<<8)
#define	KINFO_RT		(1<<8)
#define	KINFO_VNODE		(2<<8)
#define	KINFO_FILE		(3<<8)
#define	KINFO_METER		(4<<8)
#define	KINFO_LOADAVG		(5<<8)
#define	KINFO_CLOCKRATE		(6<<8)
#define KINFO_SYSINFO           (101 << 8)

#ifdef __i386__
#define	COMPAT_SYSINFO_OSTYPE	"BSD/OS"
#define COMPAT_SYSINFO_MACHINE	"i386"

#define SI_STR_LEN	20
char sysinfo_str[SI_STR_LEN] = COMPAT_SYSINFO_OSTYPE "\0"
			 COMPAT_SYSINFO_MACHINE
			 /* "\0" COMPAT_SYSINFO_RELEASE */;

#define SI_OFF_OSTYPE	sizeof(struct sysinfo)
#define SI_OFF_MACHINE	(SI_OFF_OSTYPE+sizeof(COMPAT_SYSINFO_OSTYPE))
#define SI_OFF_RELEASE	(SI_OFF_MACHINE+sizeof(COMPAT_SYSINFO_MACHINE))
#define SI_STROFF_RELEASE	(sizeof(COMPAT_SYSINFO_OSTYPE) + \
				 sizeof(COMPAT_SYSINFO_MACHINE))

#define SI_NEEDED	sizeof(struct sysinfo) + sizeof(sysinfo_str)
#endif /* __i386__ */

int
compat_getkerninfo_getkerninfo(p, uap, retval)
	struct proc *p;
	register struct compat_getkerninfo_getkerninfo_args /* {
		syscallarg(int) op;
		syscallarg(char *) where;
		syscallarg(int *) size;
		syscallarg(int) arg;
	} */ *uap;
	register_t *retval;
{
	int error, name[5];
	size_t size;

	if (SCARG(uap, size) && (error = copyin((caddr_t)SCARG(uap, size),
	    (caddr_t)&size, sizeof(size))))
		return (error);

	switch (SCARG(uap, op) & 0xff00) {

	case KINFO_RT:
		name[0] = PF_ROUTE;
		name[1] = 0;
		name[2] = (SCARG(uap, op) & 0xff0000) >> 16;
		name[3] = SCARG(uap, op) & 0xff;
		name[4] = SCARG(uap, arg);
		error =
		    net_sysctl(name, 5, SCARG(uap, where), &size, NULL, 0, p);
		break;

	case KINFO_VNODE:
		name[0] = KERN_VNODE;
		error =
		    kern_sysctl(name, 1, SCARG(uap, where), &size, NULL, 0, p);
		break;

	case KINFO_PROC:
		name[0] = KERN_PROC;
		name[1] = SCARG(uap, op) & 0xff;
		name[2] = SCARG(uap, arg);
		error =
		    kern_sysctl(name, 3, SCARG(uap, where), &size, NULL, 0, p);
		break;

	case KINFO_FILE:
		name[0] = KERN_FILE_COMPAT;
		error =
		    kern_sysctl(name, 1, SCARG(uap, where), &size, NULL, 0, p);
		break;

	case KINFO_METER:
		name[0] = VM_TOTAL;
		error =
		    vm_sysctl(name, 1, SCARG(uap, where), &size, NULL, 0, p);
		break;

	case KINFO_LOADAVG:
		name[0] = VM_LOADAVG;
		error =
		    vm_sysctl(name, 1, SCARG(uap, where), &size, NULL, 0, p);
		break;

#ifdef __i386__
	case KINFO_SYSINFO: {
		struct sysinfo si;
		extern char osrelease[];

		/* Fill in just enough to keep the 1.1 libc uname() happy */
		bzero(&si, sizeof(si));
		si.sys_ostype = (char *)SI_OFF_OSTYPE;
		si.sys_osrelease = (char *)SI_OFF_RELEASE;
		si.sys_machine = (char *)SI_OFF_MACHINE;
		(void) strncpy(sysinfo_str + SI_STROFF_RELEASE,
		    osrelease, SI_STR_LEN - SI_STROFF_RELEASE - 1);

		/*
		 * As with the above compat calls this does not exactly 
		 * duplicate getkerninfo(): if a destination is given and
		 * the length isn't big enough return ENOMEM (the 1.1
		 * getsysinfo() would do a partial copy, set size copied 
		 * and return the length actually needed).
		 */
		if (SCARG(uap, where) == NULL) {
			error = 0;
			size = SI_NEEDED;
			break;
		}
		if (size < SI_NEEDED) {
			error = ENOMEM;
			break;
		}
		size = SI_NEEDED;

		/* Copy out the sysinfo and strings */
		if ((error =
		    copyout(&si, SCARG(uap, where), sizeof(si))) != 0)
			break;
		error = copyout(sysinfo_str,
		    SCARG(uap, where) + sizeof(si), sizeof(sysinfo_str));
		break;
	}
#endif

	case KINFO_CLOCKRATE:
		name[0] = KERN_CLOCKRATE;
		error =
		    kern_sysctl(name, 1, SCARG(uap, where), &size, NULL, 0, p);
		break;

	default:
		return (EOPNOTSUPP);
	}
	if (error)
		return (error);
	*retval = size;
	if (SCARG(uap, size))
		error = copyout((caddr_t)&size, (caddr_t)SCARG(uap, size),
		    sizeof(size));
	return (error);
}
#endif /* COMPAT_GETKERNINFO */
