/*-
 * Copyright (c) 1996 Berkeley Software Design, Inc. All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 * BSDI sysv_msg.c,v 1.3 1996/07/22 21:39:08 mdickson Exp
 */

/*
 * Implementation of SVID messages
 *
 * Author:  Daniel Boulet
 *
 * Copyright 1993 Daniel Boulet and RTMX Inc.
 *
 * This system call was implemented by Daniel Boulet under contract from RTMX.
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * Redistribution in binary form may occur without any restrictions.
 * Obviously, it would be nice if you gave credit where credit is due
 * but requiring it would be too onerous.
 *
 * This software is provided ``AS IS'' without any warranties of any kind.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/proc.h>
#include <sys/ucred.h>
#include <sys/mount.h>
#include <sys/semaphore.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include "syscallargs.h"

static void msg_freehdr __P((struct msg * msghdr));

static int nfree_msgmaps;	/* # of free map entries */
static short free_msgmaps;	/* linked list head: free map entries */
static struct msg *free_msghdrs;/* list of free msg headers */
char   *msgpool;		/* MSGMAX byte long msg buffer pool */
struct msgmap *msgmaps;		/* MSGSEG msgmap structures */
struct msg *msghdrs;		/* MSGTQL msg headers */
struct msqid_ds *msqids;	/* MSGMNI msqid_ds struct's */

static oncecontrol_t once = ONCECONTROL_INITIALIZER;

static void
msginit()
{
	register int i;
	size_t size;

	size = msginfo.msgmax;
	msgpool = malloc(size, M_MSG, M_WAITOK);
	bzero(msgpool, size);
	size = sizeof(struct msgmap) * msginfo.msgseg;
	msgmaps = malloc(size, M_MSG, M_WAITOK);
	bzero(msgmaps, size);
	size = sizeof(struct msg) * msginfo.msgtql;
	msghdrs = malloc(size, M_MSG, M_WAITOK);
	bzero(msghdrs, size);
	size = sizeof(struct msqid_ds) * msginfo.msgmni;
	msqids = malloc(size, M_MSG, M_WAITOK);
	bzero(msqids, size);

	/* msginfo.msgssz should be a power of two for efficiency reasons. It
	 * is also pretty silly if msginfo.msgssz is less than 8 or greater
	 * than about 256 so ... */

	i = 8;
	while (i < 1024 && i != msginfo.msgssz)
		i <<= 1;
	if (i != msginfo.msgssz) {
		printf("msginfo.msgssz=%d (0x%x)\n", msginfo.msgssz,
		    msginfo.msgssz);
		panic("msginfo.msgssz not a small power of 2");
	}
	if (msginfo.msgseg > 32767) {
		printf("msginfo.msgseg=%d\n", msginfo.msgseg);
		panic("msginfo.msgseg > 32767");
	}
	if (msgmaps == NULL)
		panic("msgmaps is NULL");

	for (i = 0; i < msginfo.msgseg; i++) {
		if (i > 0)
			msgmaps[i - 1].next = i;
		msgmaps[i].next = -1;	/* implies entry is available */
	}
	free_msgmaps = 0;
	nfree_msgmaps = msginfo.msgseg;

	if (msghdrs == NULL)
		panic("msghdrs is NULL");

	for (i = 0; i < msginfo.msgtql; i++) {
		msghdrs[i].msg_type = 0;
		if (i > 0)
			msghdrs[i - 1].msg_next = &msghdrs[i];
		msghdrs[i].msg_next = NULL;
	}
	free_msghdrs = &msghdrs[0];

	if (msqids == NULL)
		panic("msqids is NULL");

	for (i = 0; i < msginfo.msgmni; i++) {
		msqids[i].msg_qbytes = 0;	/* implies entry is available */
		msqids[i].msg_perm.seq = 0;	/* reset to a known value */
	}
}

static void
msg_freehdr(msghdr)
	struct msg *msghdr;
{
	while (msghdr->msg_ts > 0) {
		short   next;
		if (msghdr->msg_spot < 0 || msghdr->msg_spot >= msginfo.msgseg)
			panic("msghdr->msg_spot out of range");
		next = msgmaps[msghdr->msg_spot].next;
		msgmaps[msghdr->msg_spot].next = free_msgmaps;
		free_msgmaps = msghdr->msg_spot;
		nfree_msgmaps++;
		msghdr->msg_spot = next;
		if (msghdr->msg_ts >= msginfo.msgssz)
			msghdr->msg_ts -= msginfo.msgssz;
		else
			msghdr->msg_ts = 0;
	}
	if (msghdr->msg_spot != -1)
		panic("msghdr->msg_spot != -1");
	msghdr->msg_next = free_msghdrs;
	free_msghdrs = msghdr;
}

int
msgctl(p, uap, retval)
	struct proc *p;
	register struct msgctl_args *uap;
	int    *retval;
{
	int msqid = SCARG(uap, msqid);
	int cmd = SCARG(uap, cmd);
	struct msqid_ds *user_msqptr = (struct msqid_ds *) SCARG(uap, buf);
	struct ucred *cred = p->p_ucred;
	int rval, eval;
	struct msqid_ds msqbuf;
	register struct msqid_ds *msqptr;
	struct msg *msghdr;

	msqid = IPCID_TO_IX(msqid);
	if (msqid < 0 || msqid >= msginfo.msgmni)
		return (EINVAL);

	onceinit(&once, msginit);

	msqptr = &msqids[msqid];
	if (msqptr->msg_qbytes == 0)
		return (EINVAL);

	if (msqptr->msg_perm.seq != IPCID_TO_SEQ(SCARG(uap, msqid)))
		return (EINVAL);

	eval = 0;
	rval = 0;

	switch (cmd) {

	case IPC_RMID:
		if ((eval = ipcperm(cred, &msqptr->msg_perm, IPC_M)) != 0)
			return (eval);

		/* Free the message headers */
		msghdr = msqptr->msg_first;
		while (msghdr != NULL) {
			struct msg *msghdr_tmp;

			/* Free the segments of each message */
			msqptr->msg_cbytes -= msghdr->msg_ts;
			msqptr->msg_qnum--;
			msghdr_tmp = msghdr;
			msghdr = msghdr->msg_next;
			msg_freehdr(msghdr_tmp);
		}

		if (msqptr->msg_cbytes != 0)
			panic("msg_cbytes is screwed up");
		if (msqptr->msg_qnum != 0)
			panic("msg_qnum is screwed up");

		msqptr->msg_qbytes = 0;	/* Mark it as free */

		wakeup(msqptr);
		break;

	case IPC_SET:
		if ((eval = ipcperm(cred, &msqptr->msg_perm, IPC_M)) != 0)
			return (eval);
		if ((eval = copyin(user_msqptr, &msqbuf, sizeof(msqbuf))) != 0)
			return (eval);
		if (msqbuf.msg_qbytes > msqptr->msg_qbytes && cred->cr_uid != 0)
			return (EPERM);
		if (msqbuf.msg_qbytes > msginfo.msgmnb)
			/* silently restrict qbytes to system limit */
			msqbuf.msg_qbytes = msginfo.msgmnb;

		if (msqbuf.msg_qbytes == 0)
			return (EINVAL);	/* non-standard errno! */

		/* change the owner */
		msqptr->msg_perm.uid = msqbuf.msg_perm.uid;
		msqptr->msg_perm.gid = msqbuf.msg_perm.gid;

		msqptr->msg_perm.mode = (msqptr->msg_perm.mode & ~0777) |
		    (msqbuf.msg_perm.mode & 0777);
		msqptr->msg_qbytes = msqbuf.msg_qbytes;
		msqptr->msg_ctime = time.tv_sec;
		break;

	case IPC_STAT:
		if ((eval = ipcperm(cred, &msqptr->msg_perm, IPC_R)) != 0)
			return (eval);

		eval = copyout((caddr_t) msqptr, user_msqptr,
		    sizeof(struct msqid_ds));
		break;

	default:
		return (EINVAL);
	}

	if (eval == 0)
		*retval = rval;
	return (eval);
}

int
msgget(p, uap, retval)
	struct proc *p;
	register struct msgget_args *uap;
	int    *retval;
{
	int msqid, eval;
	int key = SCARG(uap, key);
	int msgflg = SCARG(uap, msgflg);
	struct ucred *cred = p->p_ucred;
	register struct msqid_ds *msqptr = NULL;

	onceinit(&once, msginit);

	if (key != IPC_PRIVATE) {
		for (msqid = 0; msqid < msginfo.msgmni; msqid++) {
			msqptr = &msqids[msqid];
			if (msqptr->msg_qbytes != 0 && 
			    msqptr->msg_perm.key == key)
				break;
		}
		if (msqid < msginfo.msgmni) {
			if ((msgflg & IPC_CREAT) && (msgflg & IPC_EXCL))
				return (EEXIST);

			eval = ipcperm(cred, &msqptr->msg_perm, msgflg & 0700);
			if (eval != 0)
				return (eval);

			goto found;
		}
	}
	if (key == IPC_PRIVATE || (msgflg & IPC_CREAT)) {
		for (msqid = 0; msqid < msginfo.msgmni; msqid++) {

			/* Look for an unallocated and unlocked msqid_ds.
			 * msqid_ds's can be locked by msgsnd or msgrcv while
			 * they are copying the message in/out.  We can't
			 * re-use the entry until they release it. */

			msqptr = &msqids[msqid];
			if (msqptr->msg_qbytes == 0 &&
			    (msqptr->msg_perm.mode & MSG_LOCKED) == 0)
				break;
		}
		if (msqid == msginfo.msgmni)
			return (ENOSPC);

		msqptr->msg_perm.key = key;
		msqptr->msg_perm.cuid = cred->cr_uid;
		msqptr->msg_perm.uid = cred->cr_uid;
		msqptr->msg_perm.cgid = cred->cr_gid;
		msqptr->msg_perm.gid = cred->cr_gid;
		msqptr->msg_perm.mode = (msgflg & 0777);

		/* Make sure that the returned msqid is unique */
		msqptr->msg_perm.seq++;
		msqptr->msg_first = NULL;
		msqptr->msg_last = NULL;
		msqptr->msg_cbytes = 0;
		msqptr->msg_qnum = 0;
		msqptr->msg_qbytes = msginfo.msgmnb;
		msqptr->msg_lspid = 0;
		msqptr->msg_lrpid = 0;
		msqptr->msg_stime = 0;
		msqptr->msg_rtime = 0;
		msqptr->msg_ctime = time.tv_sec;
	} else
		return (ENOENT);

found:
	/* Construct the unique msqid */
	*retval = IXSEQ_TO_IPCID(msqid, msqptr->msg_perm);
	return (0);
}

int
msgsnd(p, uap, retval)
	struct proc *p;
	register struct msgsnd_args *uap;
	int    *retval;
{
	int     msqid = SCARG(uap, msqid);
	void   *user_msgp = SCARG(uap, msgp);
	size_t  msgsz = SCARG(uap, msgsz);
	int     msgflg = SCARG(uap, msgflg);
	int     segs_needed, eval;
	struct ucred *cred = p->p_ucred;
	register struct msqid_ds *msqptr;
	register struct msg *msghdr;
	short   next;

	msqid = IPCID_TO_IX(msqid);
	if (msqid < 0 || msqid >= msginfo.msgmni)
		return (EINVAL);

	onceinit(&once, msginit);

	msqptr = &msqids[msqid];
	if (msqptr->msg_qbytes == 0)
		return (EINVAL);

	if (msqptr->msg_perm.seq != IPCID_TO_SEQ(SCARG(uap, msqid)))
		return (EINVAL);

	if ((eval = ipcperm(cred, &msqptr->msg_perm, IPC_W)) != 0)
		return (eval);

	segs_needed = (msgsz + msginfo.msgssz - 1) / msginfo.msgssz;

	for (;;) {
		int need_more_resources;
		int we_own_it;

		/* check msgsz (inside this loop in case msg_qbytes 
		 * changes while we sleep) */

		if (msgsz > msqptr->msg_qbytes)
			return (EINVAL);

		need_more_resources = 0;
		if ((msqptr->msg_perm.mode & MSG_LOCKED) ||
		    (msgsz + msqptr->msg_cbytes > msqptr->msg_qbytes) ||
		    (segs_needed > nfree_msgmaps) ||
		    (free_msghdrs == NULL))
			need_more_resources = 1;

		if (need_more_resources == 0) 
			break;

		/* We need to get or wait on a resource */
		if ((msgflg & IPC_NOWAIT) != 0)
			return (EAGAIN);

		we_own_it = 0;
		if ((msqptr->msg_perm.mode & MSG_LOCKED) == 0) {
			/* Later arrivals  wait for our request */
			msqptr->msg_perm.mode |= MSG_LOCKED;
			we_own_it = 1;
		}

		eval = tsleep(msqptr, (PZERO - 4) | PCATCH, "msgwait", 0);
		if (we_own_it)
			msqptr->msg_perm.mode &= ~MSG_LOCKED;

		if (eval != 0)
			return (EINTR);

		/* Make sure that the msq queue still exists */
		if (msqptr->msg_qbytes == 0) 
			return (EIDRM);
	}

	/* We have the resources that we need. Make sure! */

	if (msqptr->msg_perm.mode & MSG_LOCKED)
		panic("msg_perm.mode & MSG_LOCKED");
	if (segs_needed > nfree_msgmaps)
		panic("segs_needed > nfree_msgmaps");
	if (msgsz + msqptr->msg_cbytes > msqptr->msg_qbytes)
		panic("msgsz + msg_cbytes > msg_qbytes");
	if (free_msghdrs == NULL)
		panic("no more msghdrs");

	/* Re-lock the msqid_ds in case we page-fault when copying in the
	 * message */

	if ((msqptr->msg_perm.mode & MSG_LOCKED) != 0)
		panic("msqid_ds is already locked");
	msqptr->msg_perm.mode |= MSG_LOCKED;

	/* Allocate a message header */

	msghdr = free_msghdrs;
	free_msghdrs = msghdr->msg_next;
	msghdr->msg_spot = -1;
	msghdr->msg_ts = msgsz;

	/* Allocate space for the message */

	while (segs_needed > 0) {
		if (nfree_msgmaps <= 0)
			panic("not enough msgmaps");
		if (free_msgmaps == -1)
			panic("nil free_msgmaps");
		next = free_msgmaps;
		if (next <= -1)
			panic("next too low #1");
		if (next >= msginfo.msgseg)
			panic("next out of range #1");

		free_msgmaps = msgmaps[next].next;
		nfree_msgmaps--;
		msgmaps[next].next = msghdr->msg_spot;
		msghdr->msg_spot = next;
		segs_needed--;
	}

	/* Copy in the message type */

	if ((eval = copyin(user_msgp, &msghdr->msg_type, 
		           sizeof(msghdr->msg_type))) != 0) {
		msg_freehdr(msghdr);
		msqptr->msg_perm.mode &= ~MSG_LOCKED;
		wakeup(msqptr);
		return (eval);
	}
	user_msgp += sizeof(msghdr->msg_type);

	/* Validate the message type */
	if (msghdr->msg_type < 1) {
		msg_freehdr(msghdr);
		msqptr->msg_perm.mode &= ~MSG_LOCKED;
		wakeup(msqptr);
		return (EINVAL);
	}

	/* Copy in the message body */
	next = msghdr->msg_spot;
	while (msgsz > 0) {
		size_t  tlen;
		if (msgsz > msginfo.msgssz)
			tlen = msginfo.msgssz;
		else
			tlen = msgsz;
		if (next <= -1)
			panic("next too low #2");
		if (next >= msginfo.msgseg)
			panic("next out of range #2");
		if ((eval = copyin(user_msgp, &msgpool[next * msginfo.msgssz],
			    tlen)) != 0) {
			msg_freehdr(msghdr);
			msqptr->msg_perm.mode &= ~MSG_LOCKED;
			wakeup(msqptr);
			return (eval);
		}
		msgsz -= tlen;
		user_msgp += tlen;
		next = msgmaps[next].next;
	}

	if (next != -1)
		panic("didn't use all the msg segments");

	/* We've got the message.  Unlock the msqid_ds. */
	msqptr->msg_perm.mode &= ~MSG_LOCKED;

	/* Make sure that the msqid_ds is still allocated. */
	if (msqptr->msg_qbytes == 0) {
		msg_freehdr(msghdr);
		wakeup(msqptr);
		return (EIDRM);
	}
	/* Put the message into the queue */

	if (msqptr->msg_first == NULL) {
		msqptr->msg_first = msghdr;
		msqptr->msg_last = msghdr;
	} else {
		msqptr->msg_last->msg_next = msghdr;
		msqptr->msg_last = msghdr;
	}
	msqptr->msg_last->msg_next = NULL;

	msqptr->msg_cbytes += msghdr->msg_ts;
	msqptr->msg_qnum++;
	msqptr->msg_lspid = p->p_pid;
	msqptr->msg_stime = time.tv_sec;

	wakeup(msqptr);
	*retval = 0;
	return (0);
}

int
msgrcv(p, uap, retval)
	struct proc *p;
	register struct msgrcv_args *uap;
	int *retval;
{
	int msqid = SCARG(uap, msqid);
	void *user_msgp = SCARG(uap, msgp);
	size_t msgsz = SCARG(uap, msgsz);
	long msgtyp = SCARG(uap, msgtyp);
	int msgflg = SCARG(uap, msgflg);
	size_t len;
	struct ucred *cred = p->p_ucred;
	register struct msqid_ds *msqptr;
	register struct msg *msghdr;
	int eval;
	short next;

	msqid = IPCID_TO_IX(msqid);
	if (msqid < 0 || msqid >= msginfo.msgmni)
		return (EINVAL);

	onceinit(&once, msginit);

	msqptr = &msqids[msqid];
	if (msqptr->msg_qbytes == 0)
		return (EINVAL);

	if (msqptr->msg_perm.seq != IPCID_TO_SEQ(SCARG(uap, msqid)))
		return (EINVAL);

	if ((eval = ipcperm(cred, &msqptr->msg_perm, IPC_R)) != 0)
		return (eval);

	msghdr = NULL;
	while (msghdr == NULL) {
		if (msgtyp == 0) {
			msghdr = msqptr->msg_first;
			if (msghdr != NULL) {
				if (msgsz < msghdr->msg_ts &&
				    (msgflg & MSG_NOERROR) == 0)
					return (E2BIG);

				if (msqptr->msg_first == msqptr->msg_last) {
					msqptr->msg_first = NULL;
					msqptr->msg_last = NULL;
				} else {
					msqptr->msg_first = msghdr->msg_next;
					if (msqptr->msg_first == NULL)
						panic("msg_first/last screwed up #1");
				}
			}
		} else {
			struct msg *previous;
			struct msg **prev;

			previous = NULL;
			prev = &(msqptr->msg_first);
			while ((msghdr = *prev) != NULL) {
				/* Is this message's type an exact match or is
				 * this message's type less than or equal to
				 * the absolute value of a negative msgtyp?
				 * Note that the second half of this test can
				 * NEVER be true if msgtyp is positive since
				 * msg_type is always positive! */

				if (msgtyp == msghdr->msg_type ||
				    msghdr->msg_type <= -msgtyp) {
					if (msgsz < msghdr->msg_ts &&
					    (msgflg & MSG_NOERROR) == 0)
						return (E2BIG);

					*prev = msghdr->msg_next;
					if (msghdr == msqptr->msg_last) {
						if (previous == NULL) {
							if (prev !=
							    &msqptr->msg_first)
								panic("msg_first/last screwed up #2");
							msqptr->msg_first =
							    NULL;
							msqptr->msg_last =
							    NULL;
						} else {
							if (prev ==
							    &msqptr->msg_first)
								panic("msg_first/last screwed up #3");
							msqptr->msg_last =
							    previous;
						}
					}
					break;
				}
				previous = msghdr;
				prev = &(msghdr->msg_next);
			}
		}

		/* We've either extracted the msghdr for the appropriate
		 * message or there isn't one. If there is one then bail out
		 * of this loop. */

		if (msghdr != NULL)
			break;

		/* Hmph!  No message found.  Does the user want to wait? */

		if ((msgflg & IPC_NOWAIT) != 0)
			return (ENOMSG);

		/* Wait for something to happen */
		eval = tsleep(msqptr, (PZERO - 4) | PCATCH, "msgwait", 0);
		if (eval != 0)
			return (EINTR);

		/* Make sure that the msq queue still exists */
		if (msqptr->msg_qbytes == 0 ||
		    msqptr->msg_perm.seq != IPCID_TO_SEQ(SCARG(uap, msqid))) 
			return (EIDRM);
	}

	/* Return the message to the user.
	 * 
	 * First, do the bookkeeping (before we risk being interrupted). */

	msqptr->msg_cbytes -= msghdr->msg_ts;
	msqptr->msg_qnum--;
	msqptr->msg_lrpid = p->p_pid;
	msqptr->msg_rtime = time.tv_sec;

	/* Make msgsz the actual amount that we'll be returning. Note that
	 * this effectively truncates the message if it is too long (since
	 * msgsz is never increased). */
	if (msgsz > msghdr->msg_ts)
		msgsz = msghdr->msg_ts;

	/* Return the type to the user. */

	eval = copyout((caddr_t) & (msghdr->msg_type), user_msgp,
	    sizeof(msghdr->msg_type));
	if (eval != 0) {
		msg_freehdr(msghdr);
		wakeup(msqptr);
		return (eval);
	}
	user_msgp += sizeof(msghdr->msg_type);

	/* Return the segments to the user */

	next = msghdr->msg_spot;
	for (len = 0; len < msgsz; len += msginfo.msgssz) {
		size_t  tlen;

		if (msgsz > msginfo.msgssz)
			tlen = msginfo.msgssz;
		else
			tlen = msgsz;
		if (next <= -1)
			panic("next too low #3");
		if (next >= msginfo.msgseg)
			panic("next out of range #3");
		eval = copyout((caddr_t) & msgpool[next * msginfo.msgssz],
		    user_msgp, tlen);
		if (eval != 0) {
			msg_freehdr(msghdr);
			wakeup(msqptr);
			return (eval);
		}
		user_msgp += tlen;
		next = msgmaps[next].next;
	}

	/* Done, return the actual number of bytes copied out. */

	msg_freehdr(msghdr);
	wakeup(msqptr);
	*retval = msgsz;
	return (0);
}
