/*-
 * Copyright (c) 1996 Berkeley Software Design, Inc. All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI if_media.c,v 1.4 1996/09/13 03:18:03 ewv Exp
 */

/*
 * Utility routines called by network drivers to manage media selection
 */

#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/sockio.h>

#include <net/if.h>
#include <net/if_media.h>

/*
 * Initialize the ifmedia structure in a driver softc
 */
void
ifmedia_init(mp, mask, change, status)
	struct ifmedia *mp;		/* Media structure to initialize */
	int mask;			/* Media options don't care mask */
	ifm_change_cb_t change;		/* Media change callback */
	ifm_stat_cb_t status;		/* Media status callback */
{
	mp->ifm_mask = mask;
	mp->ifm_cur = NULL;
	mp->ifm_list = NULL;
	mp->ifm_change = change;
	mp->ifm_status = status;
}

/*
 * Add a single valid media option to an ifmedia chain
 */
void
ifmedia_add(mp, mword, data, aux)
	struct ifmedia *mp;	/* Media structure in softc */
	int             mword;	/* Media options word */
	int             data;	/* Data passed back to driver */
	void           *aux;	/* Data passed back to driver */
{
	struct ifmedia_entry *mep;

#ifdef DEBUG
	/* Make sure no one else has already put one of these on here */
	for (mep = mp->ifm_list; mep != NULL; mep = mep->ifm_next)
		if (mep->ifm_media == mword)
			break;
	if (mep != NULL) {
		printf("ifmedia_add: duplicate entry for %x\n", mword);
		panic("ifmedia_add dup");
	}
#endif
	/*
	 * This may be called during boot time attach; its more likely we
	 * will need to wait during media changes after the system is up.
	 * This beats panicing even if we might get stuck during attach.
	 */
	mep = malloc(sizeof(*mep), M_DEVBUF, M_WAITOK);

	mep->ifm_next = mp->ifm_list;		/* Put on list */
	mp->ifm_list = mep;
	mep->ifm_media = mword;
	mep->ifm_data = data;
	mep->ifm_aux = aux;
}

/*
 * Add a (generally static) list of media entries
 */
void
ifmedia_list_add(mp, lp, count)
	struct ifmedia *mp;		/* Media structure in softc */
	struct ifmedia_entry *lp;	/* Pointer to list of ifmedia entries */
	int count;			/* Number of entries to add */
{
	struct ifmedia_entry *mep;

	for (mep = lp; mep < &lp[count]; mep++)
		ifmedia_add(mp, mep->ifm_media, mep->ifm_data, mep->ifm_aux);
}

/*
 * Set current media (called from driver when it knows what its running)
 */
void
ifmedia_set(mp, mword)
	struct ifmedia *mp;
	int mword;
{
	struct ifmedia_entry *mep;

	for (mep = mp->ifm_list; mep != NULL; mep = mep->ifm_next)
		if ((mword & ~mp->ifm_mask) == mep->ifm_media)
			break;
	if (mep == NULL)
		panic("ifmedia_set: Invalid media selection");
	mp->ifm_media = mword;
	mp->ifm_cur = mep;
}

/*
 * Handle ioctls SIOCSIFMEDIA and SIOCGIFMEDIA
 */
int
ifmedia_ioctl(ifp, ifr, mp, cmd)
	struct ifnet *ifp;	/* Intrface, used for callbacks */
	struct ifreq *ifr;	/* Request structure from ioctl */
	struct ifmedia *mp;	/* Ifmedia structure in softc */
	u_long cmd;		/* Ioctl command */
{
	struct ifmedia_entry *mep;
	int rc = 0;
	struct ifmediareq *req;
	int cnt;
	int found_cnt;
	int *uptr;
	struct ifmedia_entry *prev_mep;
	int prev_mword;

	switch (cmd) {
	case SIOCSIFMEDIA:
		for (mep = mp->ifm_list; mep != NULL; mep = mep->ifm_next)
			if ((ifr->ifr_media & ~mp->ifm_mask) == mep->ifm_media)
				break;
		if (mep == NULL) {
			rc = ENXIO;
			break;
		}
		if (ifr->ifr_media == mp->ifm_media)	/* No change */
			break;
		prev_mep = mp->ifm_cur;
		prev_mword = mp->ifm_media;
		mp->ifm_cur = mep;
		mp->ifm_media = ifr->ifr_media;
		if ((rc = mp->ifm_change(ifp)) != 0) {
			mp->ifm_cur = prev_mep;
			mp->ifm_media = prev_mword;
		}
		break;

	case SIOCGIFMEDIA:
		req = (struct ifmediareq *)ifr;
		if (mp->ifm_cur != NULL)
			req->ifm_current = mp->ifm_media;
		else
			req->ifm_current = IFM_NONE;
		req->ifm_mask = mp->ifm_mask;
		req->ifm_active = req->ifm_current;
		req->ifm_status = 0;
		mp->ifm_status(ifp, req);
		cnt = req->ifm_count;
		uptr = req->ifm_ulist;
		mep = mp->ifm_list;
		found_cnt = 0;
		while (cnt--) {
			if (copyout((caddr_t)&mep->ifm_media, uptr,
			    sizeof(*uptr)))
				return (EFAULT);
			mep = mep->ifm_next;
			uptr++;
			found_cnt++;
		}
		while (mep != NULL) {
			if (req->ifm_count != 0)
				rc = E2BIG;
			found_cnt++;
			mep = mep->ifm_next;
		}
		req->ifm_count = found_cnt;
		break;

	default:
		rc = EINVAL;		/* Broken driver ... */
	}
	return (rc);
}
