/*-
 * Copyright (c) 1993, 1994, 1995, 1996 Berkeley Software Design, Inc.
 * All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI if_ppp.c,v 2.18.2.2 1997/03/31 20:42:42 prb Exp
 */

/*
 * This code is partially derived from CMU PPP.
 *
 * Copyright (c) 1989 Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Carnegie Mellon University.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * Copyright (c) 1997
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Daedalus Research 
 *      Group of the University of California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


/*
 * PPP protocol routines
 *
 * Most of these routines are specified in if_p2psubr.c in the p2pprotosw.
 * They should appear in order in this file.
 * Local PPP routines are at the end.
 */

/* 
 * Added SIF_* routines and an sif_queue data structure to do "semantic"
 * queueing of packets on the PPP interface.  This helps us to try out
 * interesting filtering and scheduling policies on packets and acks 
 * that are present in the PPP queue.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/proc.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/syslog.h>

#include <net/if.h>
#include <net/netisr.h>
#include <net/route.h>
#include <net/if_llc.h>
#include <net/if_dl.h>
#include <net/if_types.h>

#include <machine/cpu.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#ifdef SEMANTIC_QUEUE
#include <netinet/tcp.h>
#endif SEMANTIC_QUEUE
#endif

#ifdef NS
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif

#ifdef ISO
#include <netiso/argo_debug.h>
#include <netiso/iso.h>
#include <netiso/iso_var.h>
#include <netiso/iso_snpac.h>
#endif

#include <net/slcompress.h>
#include <net/if_p2p.h>
#include <net/if_ppp.h>
#include <net/if_pppioctl.h>
#ifdef	MULTILINK
#include <net/if_pif.h>
#endif

extern struct timeval time;
struct protoq ppp_protoq;

struct p2pcom	*ppp_ifs;	/* the list of active PPP interfaces */
int		ppp_idletimer; /* idle timer is active */


int ppp_output __P((struct ifnet *, struct mbuf *, struct sockaddr *, struct
    rtentry *));
static void ppp_idlepoll __P((void *));
static void ppp_rawintr __P((void *));

#ifdef GATEWAY
/*
 * We want to put "interactive" traffic on a high priority queue.
 * Packets with IPTOS_LOWDELAY are queued separately and sent
 * preferentially.  For now, we retain the following disgusting hack
 * to check TCP port numbers, as we might be forwarding packets
 * for a system that does not set the IP TOS field correctly.
 * To decide if that traffic is interactive, we check that
 * a) it is TCP and b) one of its ports is telnet, rlogin or ftp control.
 */
static u_short interactive_ports[8] = {
	0,	513,	0,	0,
	0,	21,	0,	23,
};
#define INTERACTIVE(p) (interactive_ports[(p) & 7] == (p))
#endif

#ifdef PPP_DEBUG
#define	dprintf(x)	if (pp->p2p_if.if_flags & IFF_DEBUG) printf x
#else
#define dprintf(x)
#endif


/*
 * Initialize the PPP protocol on an interface
 */
int
ppp_attach(pp)
	struct p2pcom *pp;
{
	struct ppp_sc *ppp;

#ifdef	DEBUG
	if (pp->p2p_private)
		panic("ppp_attach: interface already attached to protocol");
#endif

	ppp = malloc(sizeof(struct ppp_sc), M_DEVBUF, M_WAITOK);
	bzero(ppp, sizeof(*ppp));
	MGETHDR(pp->p2p_ibuf, M_WAITOK, MT_DATA);
	pp->p2p_ibuf->m_len = 0;
	pp->p2p_ibuf->m_pkthdr.len = 0;
	pp->p2p_private = ppp;
	ppp->ppp_vjc.vjc_mbuf = pp->p2p_ibuf;
	ppp->ppp_vjc.vjc_ifp = &pp->p2p_if;
	ppp->ppp_vjc.vjc_basement = PH_LINK;
	PQ_INIT(&ppp_protoq, ppp_rawintr);
	ppp->ppp_ipdefq.ifq_maxlen = IFQ_MAXLEN;
	return (0);
}

/*
 * Shutdown protocol on an interface
 */
void
ppp_detach(pp)
	struct p2pcom *pp;
{
	struct ppp_sc *ppp;

	if (pp->p2p_ibuf)
		m_freem(pp->p2p_ibuf);
	if (ppp = (struct ppp_sc *)pp->p2p_private) {
#ifdef	MULTILINK
		if (ppp_mp_detach)
			(*ppp_mp_detach)(pp, ppp);
#endif
		free(ppp, M_DEVBUF);
		pp->p2p_private = 0;
	}
#ifdef	DEBUG
	else
		printf("ppp_detach: not attached\n");
#endif
}

/*
 * Initialize the PPP protocol on an interface
 */
int
ppp_init(pp)
	struct p2pcom *pp;
{
	struct ppp_sc *ppp = (struct ppp_sc *)pp->p2p_private;
	int s;

#ifdef INET
	ppp->ppp_ctcp = malloc(sizeof(*ppp->ppp_ctcp), M_DEVBUF, M_WAITOK);
	ppp->ppp_ctcp->sls_ref = 0;
	sl_compress_init(ppp->ppp_ctcp);
#endif

	s = splimp();

	pp->p2p_prevp = &ppp_ifs;
	if (pp->p2p_next = ppp_ifs)
		ppp_ifs->p2p_prevp = &pp->p2p_next;
	ppp_ifs = pp;
	if (ppp_idletimer == 0) {
		ppp_idletimer++;
		timeout(ppp_idlepoll, 0, hz);
	}

	dprintf(("%s%d: init PPP link\n", pp->p2p_if.if_name,
	    pp->p2p_if.if_unit));

	ppp->ppp_wdrop = 1;	/* For any case... */
	splx(s);
	return (0);
}

/*
 * Shutdown protocol on an interface
 */
void
ppp_shutdown(pp)
	struct p2pcom *pp;
{
	struct ppp_sc *ppp = (struct ppp_sc *)pp->p2p_private;
	struct mbuf *m;
	int s;

	dprintf(("%s%d: shutdown PPP link\n", pp->p2p_if.if_name,
	    pp->p2p_if.if_unit));

	s = splimp();
	if (pp->p2p_next)
		pp->p2p_next->p2p_prevp = pp->p2p_prevp;
	*(pp->p2p_prevp) = pp->p2p_next;

	if (ppp->ppp_ctcp) {
		if (--ppp->ppp_ctcp->sls_ref < 0)
			free(ppp->ppp_ctcp, M_DEVBUF);
		ppp->ppp_ctcp = NULL;
	}
	ppp->ppp_flags &= ~PPP_WAITING;
	if_qflush(&ppp->ppp_ipdefq);
#ifndef SEMANTIC_QUEUE
	if_qflush(&pp->p2p_isnd);
	if_qflush(&pp->p2p_if.if_snd);
#else
	sif_qflush(&pp->p2p_sisnd);
	sif_qflush(&pp->p2p_if.sif_snd);
#endif
#ifdef	MULTILINK
	if (ppp_mp_detach)
		(*ppp_mp_detach)(pp, ppp);
#endif

	splx(s);
}

#ifdef SEMANTIC_QUEUE
int
SIF_ENQUEUE(ifq, m)
	struct sifqueue *ifq;
	struct semantic_mbuf *m;
{
	m->sm_nextpkt = 0;
	if (ifq->ifq_tail == 0)
		ifq->ifq_head = m;
	else
		ifq->ifq_tail->sm_nextpkt = m;
	ifq->ifq_tail = m;
	ifq->ifq_len++;
}

struct mbuf *
SIF_DEQUEUE(sifq)
	struct sifqueue *sifq;
{
	struct mbuf *m = 0;

	if (sifq->ifq_head && (m = sifq->ifq_head->mb)) {
		struct semantic_mbuf *p = sifq->ifq_head;
		if ((sifq->ifq_head = sifq->ifq_head->sm_nextpkt) == 0)
		        sifq->ifq_tail = 0; 
		m->m_nextpkt = 0;
		sifq->ifq_len--;
		free(p, M_PCB);
	}
	return m;
}

int
SIF_REMOVE(sifq, sm, prev)
	struct sifqueue *sifq;
	struct semantic_mbuf *sm, *prev;
{
	if (sifq->ifq_head == sm)
		SIF_DEQUEUE(sifq);
	else {
		prev->sm_nextpkt = sm->sm_nextpkt;
		if (sifq->ifq_tail == sm)
			sifq->ifq_tail = prev;
		m_freem(sm->mb);
		sifq->ifq_len--;
	}
}

#endif

/*
 * Output a paket to point-to-point interface with PPP encapsulation
 */
int
ppp_output(ifp, m, dst, rt)
	register struct ifnet *ifp;
	struct mbuf *m;
	struct sockaddr *dst;
	struct rtentry *rt;
{
	u_short type;
	int hl;
	int s;
	struct p2pcom *pp = (struct p2pcom *) ifp;
	struct ppp_sc *ppp = (struct ppp_sc *)pp->p2p_private;
	struct ppp_header *ph;
#ifndef SEMANTIC_QUEUE
	struct ifqueue *outq;
#else
	struct sifqueue *outq;
	struct semantic_mbuf *sm;
#endif
	struct ip *ip;
	u_char *p;
	struct mbuf *m0;
	struct slcompress_buf cb;
#ifdef	MULTILINK
	u_long	seq;
#endif

	/*
	 * IFF_RUNNING or PPP_WAITING is required to send a packet out.
	 * IFF_RUNNING is set when starting up LCP by the ppp daemon.
	 * PPP_WAITING is set by the PPPIOCWAIT ioctl
	 * IFF_UP is also required if this is not a packet from
	 * the raw link layer.
	 */
	if (((ifp->if_flags & IFF_UP) == 0 && dst->sa_family != AF_LINK) ||
	    (ifp->if_flags & IFF_RUNNING) == 0 &&
	    (ppp->ppp_flags & PPP_WAITING) == 0) {
		m_freem(m);
		return (ENETDOWN);
	}

	cb.chlen = 0;
	cb.uhlen = 0;
	cb.bpflen = pp->p2p_if.if_bpf ? SLC_BPFHDRLEN : 0;

#ifndef SEMANTIC_QUEUE
	outq = &ifp->if_snd;
#else
	outq = &ifp->sif_snd;
#endif
	switch (dst->sa_family) {
#ifdef INET
	case AF_INET:
		if (!(ppp->ppp_flags & PPP_IPOKAY)) {
			/*
			 * It is not okay to recieve IP packets.
			 * If REJECT is not turned on then we can at
			 * least try to put them on the deferred queue.
			 */
			if (ppp->ppp_flags & PPP_IPREJECT) {
				m_freem(m);
				return (EAFNOSUPPORT);
			}

			/* If someone is waiting queue to ipdefq */
			s = splimp();
			if (ppp->ppp_wdrop || ppp->ppp_ipdefq.ifq_len > 0) {
				m0 = m;
				m = 0;

        			if (IF_QFULL(&ppp->ppp_ipdefq))
					IF_DEQUEUE(&ppp->ppp_ipdefq, m);
				IF_ENQUEUE(&ppp->ppp_ipdefq, m0);
			}

			/* Wake up the daemon */
			if (ppp->ppp_wdrop) {
				wakeup((caddr_t) &ppp->ppp_wdrop);
				ppp->ppp_wdrop = 0;
			}
			splx(s);
			if (m)
				m_freem(m);
			return (0);	/* simply drop on the floor */
		}

		/*
		 * Compute how much space we will need at the front
		 * of the packet.  Need to leave space if we have
		 * a BPF header as well.
		 * We know that all packets will have a compressable
		 * protocol field at this point.
		 */
		if (cb.bpflen)
			cb.lllen = sizeof(struct ppp_header);
		else {
			cb.lllen = (ppp->ppp_flags & PPP_ACFC) ? 0 : 2;
			cb.lllen += (ppp->ppp_flags & PPP_PFC) ? 1 : 2;
		}
		if (ppp->ppp_mp)
			cb.lllen += (ppp->ppp_flags & PPP_SND_SSEQ) ? 4 : 6;

		type = PPP_IP;
		ip = mtod(m, struct ip *);

		/*
		 * Push "interactive" packets
		 * to the priority queue
		 */
		if (IPTOS_FASTQ(ip) && ifp->if_flags & IFF_FASTQ) {
#ifndef SEMANTIC_QUEUE
			outq = &pp->p2p_isnd;
#else 
			outq = &pp->p2p_sisnd;
#endif SEMANTIC_QUEUE 
		}
#ifdef GATEWAY
		else if (ip->ip_p == IPPROTO_TCP && ip->ip_off == 0 &&
			 (ppp->ppp_flags & PPP_FTEL)) {

			/* get src & dst port numbers */
			hl = ntohl(((int *)ip)[ip->ip_hl]);
			if (INTERACTIVE(hl & 0xffff) || INTERACTIVE(hl >> 16)){
#ifndef SEMANTIC_QUEUE
				outq = &pp->p2p_isnd;
#else 
				outq = &pp->p2p_sisnd;
#endif SEMANTIC_QUEUE
			}
		}
#endif

#ifdef SEMANTIC_QUEUE
		/* Get relevant TCP/IP fields and initialize semhdr */
		sm = (struct semantic_mbuf *)
			malloc(sizeof(struct semantic_mbuf), M_PCB, M_NOWAIT);
		if (sm)
			sm->sh.proto = ip->ip_p;
		if (sm && ip->ip_p == IPPROTO_TCP) {
				struct tcphdr *th = mtod(m, struct tcphdr *);
				/* XXX may need a m_pullup()... */
				sm->sh.src = ip->ip_src.s_addr;
				sm->sh.dst = ip->ip_dst.s_addr;
				sm->sh.sport = th->th_sport;
				sm->sh.dport = th->th_dport;
				sm->sh.flags = th->th_flags;
				sm->sh.ack = th->th_ack;
				sm->sh.size = ntohs(ip->ip_len)-(ip->ip_hl<<2) 
					- (th->th_off<<2);
		}
#endif
		/*
		 * Handle TCP compression here
		 */
		if (ip->ip_p == IPPROTO_TCP && (ppp->ppp_flags & PPP_TCPC)) {
			switch (sl_compress_tcp(&m, ip, ppp->ppp_ctcp,
			    SL_COMP_NOCID, &cb)) {
			case 0:
				return (ENOBUFS);
			case TYPE_COMPRESSED_TCP:
				type = PPP_VJC;
				break;
			case TYPE_UNCOMPRESSED_TCP:
				type = PPP_VJNC;
				break;
			case TYPE_IP:
				break;
			default:
				panic("Unknown type returned from sl_compress");
				break;
			}
		} else {
			if (cb.bpflen)
				hl = SLC_BPFHDR;
			else
				hl = cb.lllen;
			M_PREPEND(m, hl, M_DONTWAIT);
			if (m == 0)
				return (ENOBUFS);
		}
		break;
#endif
	case AF_LINK:
		/*
		 * Raw link level packets always take a full header.
		 */
		ph = mtod(m, struct ppp_header *);
		cb.lllen = sizeof(struct ppp_header);
		if (cb.bpflen) {
			m->m_data += cb.lllen;
			m->m_len -= cb.lllen;
			m->m_pkthdr.len -= cb.lllen;
			M_PREPEND(m, SLC_BPFHDR, M_DONTWAIT);
			if (m == 0)
				return (ENOBUFS);
			p = mtod(m, u_char *) + SLC_BPFHDRLEN;
			bcopy(ph, p, cb.lllen);
		}
#ifdef SEMANTIC_QUEUE
		sm = (struct semantic_mbuf *) malloc(sizeof(* sm),
						     M_PCB, M_NOWAIT);
		sm->sh.proto = 0; /* not tcp, that's what matters */
#endif
		break;

	default:
		printf("%s%d: af%d not supported\n", ifp->if_name,
				ifp->if_unit, dst->sa_family);
		m_freem(m);
		return (EAFNOSUPPORT);
	}

	/*
	 * Add ppp serial line header if we are not from the link layer
	 */
	if (dst->sa_family != AF_LINK) {
		/* Reset the idle time counter */
		ppp->ppp_idlecnt = 1;

		p = mtod(m, u_char *) + cb.lllen + cb.bpflen;

#ifdef	MULTILINK
		/* Insert Multilink Protocol header if needed. */
		if (ppp->ppp_mp) {
			*--p = type & 0xff;

			if (cb.bpflen || (ppp->ppp_flags & PPP_PFC) == 0
			    || (type & 0xff00))
				*--p = (type >> 8);

			seq = ppp->ppp_mp->mp_snd_seq++;
			*--p = seq & 0xff;
			if (ppp->ppp_flags & PPP_SND_SSEQ)
				*--p = ((seq >> 8) & 0x0f) |
				    ((MP_BEGIN|MP_END)<<6);
			else {
				*--p = (seq >> 8) & 0xff;
				*--p = (seq >> 16) & 0xff;
				*--p = (MP_BEGIN|MP_END)<<6;
			}
			type = PPP_MP;
		}
#endif	/* MULTILINK */

		*--p = type & 0xff;

		if (cb.bpflen || (ppp->ppp_flags & PPP_PFC) == 0
		    || (type & 0xff00))
			*--p = (type >> 8);

		if (cb.bpflen || (ppp->ppp_flags & PPP_ACFC) == 0) {
			*--p = PPP_CONTROL;
			*--p = PPP_ADDRESS;
		}
#ifdef	DEBUG
		if ((mtod(m, u_char *) + cb.lllen + cb.bpflen) - p != cb.lllen)
			panic("ppp_output: header length changed");
#endif
	}

	/*
	 * Queue packet on interface and
	 * start output if not active.
	 */
	s = splimp();
#ifndef SEMANTIC_QUEUE
	if (IF_QFULL2(&ifp->if_snd, &pp->p2p_isnd)) {
		IF_DROP(&ifp->if_snd);
		splx(s);
		m_freem(m);
		return (ENOBUFS);
	}
#else
	if (IF_QFULL2(&ifp->sif_snd, &pp->p2p_sisnd)) {
		IF_DROP(&ifp->sif_snd);
		splx(s);
		m_freem(m);
		return (ENOBUFS);
	}
#endif

	if (cb.bpflen) {
		/*
		 * At this point the mbuf still has the orignal data
		 * as well as the link level and compressed headers.
		 * We need to prepend the 3 byte "bpf" header.
		 */
		*--p = cb.chlen;
		*--p = cb.lllen;
		*--p = SLIPDIR_OUT;
		bpf_mtap(pp->p2p_if.if_bpf, m);
		hl = cb.lllen + cb.bpflen;
		if (cb.chlen && p[hl] == TYPE_UNCOMPRESSED_TCP) {
			ip->ip_p = p[hl+1];
			cb.chlen = 0;
		}
		/*
		 * Now move the link level data and compressed header up.
		 * If we are lucky we can just trim mbufs, else we copy
		 */
		if (m->m_len == SLC_BPFHDR &&
		    m->m_next && m->m_next->m_len >= cb.uhlen) {
			m->m_data += SLC_BPFHDRLEN;
			m->m_pkthdr.len -= SLC_BPFHDR - cb.chlen - cb.lllen;
			m->m_len = cb.chlen + cb.lllen;

			m->m_next->m_len -= cb.uhlen;
			m->m_pkthdr.len -= cb.uhlen;
			m->m_next->m_data += cb.uhlen;
		} else {
			hl = SLC_BPFHDR + cb.uhlen - cb.chlen - cb.lllen;
			if (hl) {
				struct mbuf *m1 = m;
				m->m_pkthdr.len -= hl;
				while (m1 && hl) {
					if (hl >= m1->m_len) {
						hl -= m1->m_len;
						m1->m_len = 0;
						if ((m1 = m1->m_next) == 0)
							panic("overran mbuf");
					} else {
						m1->m_len -= hl;
						m1->m_data += hl;
						hl = 0;
					}
				}
				hl = cb.lllen + cb.chlen;
				if (m1->m_len < hl)
					panic("IP hdr in mbuf not contiguous");
				bcopy(p + SLC_BPFHDRLEN, mtod(m1, caddr_t), hl);
			}
		}
		/*
		 * Fix up the PPP header since we did not do header compression
		 * for BPF packets.
		 * Note that when doing multilink and using BPF the second
		 * (real) protocol field (type) will not be compressed.  This
		 * will not cause a problem and only incurs a single byte
		 * overhead.
		 */
		if (dst->sa_family != AF_LINK) {
			hl = ppp->ppp_flags & (PPP_ACFC |
			    (type & 0xff00) ? 0 : PPP_PFC);
			if (hl) {
				for (m0 = m; m0 && m0->m_len == 0;
				    m0 = m0->m_next)
					;
				if (m0->m_len >= 4)
					switch (hl) {
					case PPP_ACFC|PPP_PFC:
						m0->m_data += 3;
						m0->m_len -= 3;
						break;
					case PPP_ACFC:
						m0->m_data += 2;
						m0->m_len -= 2;
						break;
					case PPP_PFC:
						m0->m_data++;
						m0->m_len--;
						m0->m_data[0] = PPP_ADDRESS;
						m0->m_data[1] = PPP_CONTROL;
						break;
					}
			}
		}
	}

	ifp->if_obytes += m->m_pkthdr.len;

#ifndef SEMANTIC_QUEUE
	IF_ENQUEUE(outq, m);
#else
	if (sm) {
		sm->mb = m;	/* header compressed */
		if (sm->sh.proto == IPPROTO_TCP && sm->sh.flags & TH_ACK)
/*		    && ppp->ppp_txflags & PPP_ACKFILTER) */
			ppp_filteracks(outq, sm);
		SIF_ENQUEUE(outq, sm); /* implicitly at tail of queue */
	}
#endif

	if ((ifp->if_flags & IFF_OACTIVE) == 0)
		(*ifp->if_start)(ifp);
	splx(s);
	return (0);
}

/*
 * Process a received PPP packet
 *
 * Make the assumption that the mbuf chain looks like:
 *
 *	MBUF0 (permanent and should not be freed or forwarded)
 *		[ 3 bytes of empty space ]
 *		Link Level
 *		VJ Compressed Header [if any]
 *
 *	MBUF1 (Uses external data of one cluster, 2K)
 *		[ m_data points 120 bytes into data ]
 *		Uncompressed Header [if any]
 *		Data
 */
int
ppp_input(pp, m)
	struct p2pcom *pp;
	struct mbuf *m;
{
	struct ppp_sc *ppp = (struct ppp_sc *)pp->p2p_private;
	struct ppp_header *ph;
	struct ifqueue *inq;
	int s;
	struct mbuf *m1;
	u_char *p;
	u_short type;
	int clen;
#ifdef	MULTILINK
	int multilink = 0;
#endif

	/*
	 * There are three ways we can be called:
	 *
	 * m == 0: The internal mbuf has been filled by ppp_hdr
	 *
	 * m != 0: ppp_hdr was called for exactly the header, m
	 *	   contains the data of the mbuf.  PPP_CSKIP bytes
	 *	   are free at the front of m.
	 *
	 * m != 0: ppp_hdr was not called. PPP_CSKIP bytes are free at
	 *	   the front of m.  This is currently not supported.
	 */

	if (pp->p2p_ibuf->m_len == 0)
		panic("ppp_hdr has not been called");

	if (m == 0)
		m = pp->p2p_ibuf;
#ifdef	MULTILINK
	else if (m == (struct mbuf *)-1) {
		m = pp->p2p_ibuf;
		multilink = 1;
	}
#endif
	else if (m->m_flags & (M_PKTHDR | M_EXT) != (M_PKTHDR | M_EXT))
		panic("ppp_input called with small or non pkthdr mbuf");
	else {
		m1 = m;
		m = pp->p2p_ibuf;
		m->m_next = m1;
		m->m_pkthdr.len = m->m_len + m1->m_pkthdr.len;
	}

	clen = 0;

	m1 = m->m_next;		/* This is the actual data */

	pp->p2p_if.if_ibytes += m->m_pkthdr.len;

	/*
	 * The ppp_header was already verified in ppp_hdr
	 */
	if (m->m_len < sizeof(struct ppp_header)) {
		if (m1)
			m_freem(m1);
		m->m_next = 0;
		m->m_len = 0;
		m->m_pkthdr.len = 0;
		return (0);
	}

	ph = mtod(m, struct ppp_header *);

	s = splimp();
	switch (type = ntohs(ph->phdr_type)) {
	default:
		if (pp->p2p_if.if_bpf) {
			m->m_pkthdr.len += SLC_BPFHDR - (clen = m->m_len);
			m->m_len = SLC_BPFHDR;
			m->m_data -= SLC_BPFHDRLEN;
			p = mtod(m, u_char *);
			p[0] = SLIPDIR_IN;
			p[1] = sizeof(struct ppp_header);
			p[2] = 0;

			bpf_mtap(pp->p2p_if.if_bpf, m);

			m->m_pkthdr.len -= SLC_BPFHDR - clen;
			m->m_len = clen;
			m->m_data += SLC_BPFHDRLEN;
		}
#ifdef	ENHANCEMENT
		if (type == PPP_LCP) {
			struct ppp_cp_hdr *cp = mtod(m1, struct ppp_cp_hdr *);
			switch (cp->cp_code) {
			case LCP_ECHO_REQ:
			case LCP_ECHO_REPL:
				ppp_cp_in(pp, m1, type);
				goto out;
			}
		}
#endif

		if (IF_QFULL(&ppp_protoq.pq_queue)) {
			IF_DROP(&ppp_protoq.pq_queue);
			m_freem(m->m_next);
		} else {
			/*
			 * We cannot forward the first mbuf in the chain
			 * as it is a permanent mbuf.  If there is data
			 * in it we copy the data, else we just copy the
			 * pkthdr.  We know the second mbuf will also have
			 * a pkthdr.
			 */
			if (m->m_len) {
				MGETHDR(m1, M_WAITOK, MT_DATA);
				if (m1 == 0) {
					m_freem(m->m_next);
					goto out;
				}
				m1->m_len = m->m_len;
				m1->m_next = m->m_next;
				m1->m_type = m->m_type;
				m1->m_flags = m->m_flags;
				m1->m_pkthdr = m->m_pkthdr;
				bcopy(m->m_data, m1->m_data, m->m_len);
			} else
				m1->m_pkthdr = m->m_pkthdr;
			PQ_ENQUEUE(&ppp_protoq, m1);
		}
		goto out;

#ifdef INET
	case PPP_IP:
	raw_ip:
		m1 = m->m_next;
		schednetisr(NETISR_IP);
		inq = &ipintrq;
		break;

	case PPP_VJC:
		if (ppp->ppp_ctcp &&
		    sl_muncompress_tcp(&m, TYPE_COMPRESSED_TCP, ppp->ppp_ctcp,
		    sizeof(struct ppp_header)) == 0)
			goto raw_ip;
		m_freem(m1);	/* m_freem handles NULL */
		goto out;

	case PPP_VJNC:
		/*
		 * We assume that sl_muncompress_tcp will always succeed
		 * on uncompressed packets.
		 */
		if (ppp->ppp_ctcp == 0) {
			m_freem(m1);
			goto out;
		}
		(void) sl_muncompress_tcp(&m, TYPE_UNCOMPRESSED_TCP,
		    ppp->ppp_ctcp, sizeof(struct ppp_header));
		goto raw_ip;
#endif
#ifdef	MULTILINK
	case PPP_MP:
		if (multilink) {
			ppp_stat.ppps_mp_mp++;
			m_freem(m1);
		}
		if ((ppp_mp_reass == NULL) || ((*ppp_mp_reass)(pp, m) < 0))
			m_freem(m1);
		goto out;
#endif	/* MULTILINK */
	}

	/*
	 * Queue the packet to incoming list
	 */
	if (IF_QFULL(inq)) {
		IF_DROP(inq);
		if (m1)
			m_freem(m1);
	} else {
		if (pp->p2p_if.if_bpf) {
			clen = m->m_len - sizeof(struct ppp_header);
			m->m_data -= SLC_BPFHDRLEN;
			m->m_pkthdr.len += SLC_BPFHDR - m->m_len;
			m->m_len = SLC_BPFHDR;

			p = mtod(m, u_char *);
			p[0] = SLIPDIR_IN;
			p[1] = sizeof(struct ppp_header);
			p[2] = clen;
			bpf_mtap(pp->p2p_if.if_bpf, m);
		}

		m1->m_pkthdr.len = m->m_pkthdr.len - m->m_len;

		IF_ENQUEUE(inq, m1);
	}
out:
	splx(s);
	m->m_next = 0;
	m->m_len = 0;
	m->m_pkthdr.len = 0;
	return (0);
}

/*
 * The PPP ioctl routine
 */
int
ppp_ioctl(pp, cmd, data)
	struct p2pcom *pp;
	int cmd;
	caddr_t data;
{
	struct ppp_sc *ppp = (struct ppp_sc *)pp->p2p_private;
	struct ppp_ioctl *io = ifr_pppioctl(data);
	int s;
	int error;
	extern struct ifnet *ppp_dropif;
	struct proc *p = curproc;	/* XXX should be parameter */

	/*
	 * We can get here via ppp_ipcp_setaddr when ipcp sets
	 * addresses from interrupt level, in which case curproc (p)
	 * is NULL.  Check suser only for calls implemented here
	 * that modify state.
	 */
	switch (cmd) {

	case PPPIOCSSTATE:
	case PPPIOCWAIT:
#ifdef	MULTILINK
	case PPPIOCSID:
	case PPPIOCCID:
	case PPPIOCSMRRU:
#endif
		if (error = suser(p->p_ucred, &p->p_acflag))
			return (error);
		break;

	case PPPIOCGSTATE:
	default:
		error = 0;
		break;
	}

	s = splimp();
	switch (cmd) {

	case PPPIOCSSTATE:
		switch (io->ppp_control & PPP_SETF) {
		case PPP_SETF:
			ppp->ppp_flags &= 0xffff0000;
			ppp->ppp_flags |= io->ppp_flags;
			break;
		case PPP_CLRF:
			ppp->ppp_flags &= ~((u_long)io->ppp_flags);
			break;
		case PPP_ADDF:
			ppp->ppp_flags |= io->ppp_flags;
			break;
		}

		switch (io->ppp_control & PPP_SETTX) {
		case PPP_SETTX:
			ppp->ppp_txcmap = io->ppp_txcmap;
			break;
		case PPP_CLRTX:
			ppp->ppp_txcmap &= ~io->ppp_txcmap;
			break;
		case PPP_ADDTX:
			ppp->ppp_txcmap |= io->ppp_txcmap;
			break;
		}

		switch (io->ppp_control & PPP_SETRX) {
		case PPP_SETRX:
			ppp->ppp_rxcmap = io->ppp_rxcmap;
			break;
		case PPP_CLRRX:
			ppp->ppp_rxcmap &= ~io->ppp_rxcmap;
			break;
		case PPP_ADDRX:
			ppp->ppp_rxcmap |= io->ppp_rxcmap;
			break;
		}

		if (io->ppp_control & PPP_SETIDLE)
			if ((ppp->ppp_idletime = io->ppp_idletime) == 0)
				ppp->ppp_idlecnt = 0;

		if (io->ppp_control & PPP_SETMTU)
			pp->p2p_if.if_mtu = io->ppp_mtu;

		/*
		 * IFF_RUNNING allows us to send/receive LCP packets.
		 */
		if (ppp->ppp_flags & PPP_RUNNING)
			pp->p2p_if.if_flags |= IFF_RUNNING;
		else
			pp->p2p_if.if_flags &= ~IFF_RUNNING;

		if ((io->ppp_control & PPP_SLCINIT) && ppp->ppp_ctcp)
			sl_compress_init(ppp->ppp_ctcp);

		if (io->ppp_control & PPP_FLUSHQ) {
#ifndef SEMANTIC_QUEUE
			if_qflush(&pp->p2p_isnd);
			if_qflush(&pp->p2p_if.if_snd);
#else
			sif_qflush(&pp->p2p_sisnd);
			sif_qflush(&pp->p2p_if.sif_snd);
#endif
		}

		if (io->ppp_control & PPP_FLUSHDQ)
			if_qflush(&ppp->ppp_ipdefq);


		if (io->ppp_control & PPP_PROCDQ)
			for (;;) {
				struct mbuf *m;
				static struct sockaddr fakedst =
					{ sizeof(struct sockaddr_in), AF_INET };

				IF_DEQUEUE(&ppp->ppp_ipdefq, m);
				if (m == 0)
					break;
				ppp_output(&pp->p2p_if, m, &fakedst, 0);
			}

		splx(s);
		return(error);

	case PPPIOCGSTATE:
		io->ppp_idletime = ppp->ppp_idletime;
		io->ppp_flags = ppp->ppp_flags;
		io->ppp_txcmap = ppp->ppp_txcmap;
		io->ppp_rxcmap = ppp->ppp_rxcmap;
		io->ppp_mtu = pp->p2p_if.if_mtu;
		break;

	case PPPIOCWAIT:	/*wait for a dropped packet (move to if_appp?)*/
		ppp->ppp_flags |= PPP_WAITING;
		if (ppp->ppp_ipdefq.ifq_len > 0)
			break;
		ppp->ppp_wdrop = 1;
		error = tsleep((caddr_t) &ppp->ppp_wdrop,
		    PCATCH|(PZERO+10), "pppwait", 0);
		break;

	default:
#ifdef	MULTILINK
		error = ppp_mp_ioctl ? (*ppp_mp_ioctl)(pp, cmd, data) : EINVAL;
#else
		error = EINVAL;
#endif
	}
	splx(s);
	return (error);
}

/*
 * Modem events handling routine
 * (0 - DCD dropped; 1 - DCD raised)
 */
void
ppp_modem(pp, dcd)
	register struct p2pcom *pp;
	int dcd;
{
	struct sockaddr_dl dl;
	struct mbuf *m;
	struct sockproto proto;

	bzero(&dl, sizeof(dl));
	dl.sdl_len = sizeof(dl);
	dl.sdl_family = AF_LINK;

	m = so_cmsg((caddr_t)&dcd, sizeof(dcd), IFS_DCD, IFT_NONE);
	if (m) {
		proto.sp_family = PF_LINK;
		proto.sp_protocol = pp->p2p_if.if_index;
		dl.sdl_index = pp->p2p_if.if_index;
		raw_input(0, m, &proto, &dl, &dl);
	}
#ifdef	DIAGNOSTIC
	else
		printf("%s%d: no available mbufs for dcd change cmsg\n",
		    pp->p2p_if.if_name, pp->p2p_if.if_unit);
#endif
}

/*
 * Put nbytes of data starting at cp into the input mbuf.
 * May be called multiple times as more data in the packet
 * arrives.
 */
int
ppp_hdr(pp, cp, nbytes)
	struct p2pcom *pp;
	u_char *cp;
	int nbytes;
{
	struct ppp_sc *ppp = (struct ppp_sc *)pp->p2p_private;

	return(vjc_hdr(&ppp->ppp_vjc, cp, nbytes));
}

#ifdef	not_used_yet
/*
 * And get/setsockopt requests.  We only need to process requiests
 * which match our interface type (IFT_PPP).
 * This is a place holder and currently not needed.
 * To use this function you need to add an entry in the p2pprotosw
 * table if_p2pproto.c.
 */int
ppp_ctloutput(op, ifp, level, optname, m)
	int op;
	struct ifnet *ifp;
	int level, optname;
	struct mbuf **m;
{
	if (level != IFT_PPP)
		return(EINVAL);

	switch(optname) {
	default:
	return(EINVAL);
	}
}
#endif


/*
 * Polling the idle time counters
 */
static void
ppp_idlepoll(arg)
	void *arg;
{
	struct ppp_sc *ppp;
	register struct p2pcom *pp, *ppn;
	int s;

	s = splimp();
	if (ppp_ifs == 0) {
		ppp_idletimer = 0;
		splx(s);
		return;
	}
	for (pp = ppp_ifs; pp; pp = ppn) {
		ppn = pp->p2p_next;
		ppp = (struct ppp_sc *)pp->p2p_private;
		if (ppp->ppp_idletime == 0)
			continue;
		if ((pp->p2p_if.if_flags & (IFF_UP|IFF_RUNNING)) !=
		    (IFF_UP|IFF_RUNNING))
			continue;
		if (ppp->ppp_idlecnt == 0)
			continue;
		if (ppp->ppp_idlecnt++ >= ppp->ppp_idletime) {
			/*
			 * Idle time expired, mark interface down
			 * and let the daemon know about it.
			 * We just drop the RUNNING bit, the user level
			 * code will down the interface.
			 */
			dprintf(("%s%d: IDLE TIMER EXPIRED\n",
				pp->p2p_if.if_name,
				pp->p2p_if.if_unit));

			pp->p2p_if.if_flags &= ~IFF_RUNNING;
			ppp->ppp_flags &= ~PPP_RUNNING;
			pp->p2p_if.if_lastchange = time;
			if_statechg(&pp->p2p_if);
		}
	}
	timeout(ppp_idlepoll, 0, hz);
	splx(s);
}

/*
 * send out any raw link level packets that were queued up.
 */
static void
ppp_rawintr(arg)
	void *arg;
{
	register struct mbuf *m;
	struct sockproto proto;
	struct sockaddr_dl dl;

	proto.sp_family = PF_LINK;
	bzero(&dl, sizeof(dl));
	dl.sdl_len = sizeof(dl);
	dl.sdl_family = AF_LINK;
	ppp_protoq.pq_wayout.pending = 0;
	for (;;) {
		IF_DEQUEUE(&ppp_protoq.pq_queue, m);
		if (m == 0)
			return;
		proto.sp_protocol = m->m_pkthdr.rcvif->if_index;
		dl.sdl_index = proto.sp_protocol;
		raw_input(m, 0, &proto, &dl, &dl);
	}
}

#ifdef SEMANTIC_QUEUE
#include <netinet/tcp_seq.h>
/* 
 * Filter "redundant" acks from queue feeding the contrained link.
 * We know for sure that sm->tcph is an ack and that outq is not empty.
 * This function must be called only at interrupt level splimp.
 */
void
ppp_filteracks(outq, new_sm)
        struct sifqueue *outq;
        struct semantic_mbuf *new_sm;
{
	struct semantic_mbuf *sm, *prev;
        struct semhdr *sh1, *sh2;
	sh1 = &(new_sm->sh);

	/* 
	 * Traverse queue, compare flows and compare ack sequence #s.
	 * Make sure not to throw away any acks that have non-zero data!
	 */
        for (sm = outq->ifq_head, prev = sm; sm != 0; ) {
	        sh2 = &(sm->sh);
		/* Ports and addrs must match, and it must be a pure ack */
		if (sh1->sport == sh2->sport && sh1->dport == sh2->dport &&
		    sh1->src == sh2->src && sh1->dst == sh2->dst &&
		    sh2->size == 0 && sh2->flags & TH_ACK) {
			if (SEQ_GT(new_sm->sh.ack, sm->sh.ack)) {
				/* Remove sm and patch up pointers */
				SIF_REMOVE(outq, sm, prev);
printf("removing ack %x from connection to %d:%d\n", sh1->src, sh1->dport);
				free(sm, M_PCB);
				sm = prev->sm_nextpkt;
				continue;
			}
		}
		prev = sm;
		sm = sm->sm_nextpkt;
	}
}

#endif
