/*
 * Copyright (c) 1988, Julian Onions <jpo@cs.nott.ac.uk>
 * Nottingham University 1987.
 *
 * This source may be freely distributed, however I would be interested
 * in any changes that are made.
 *
 * This driver takes packets off the IP i/f and hands them up to a
 * user process to have it's wicked way with. This driver has it's
 * roots in a similar driver written by Phil Cockcroft (formerly) at
 * UCL. This driver is based much more on read/write/select mode of
 * operation though.
 * 
 * if_tun.c,v 1.3 1996/08/13 02:52:46 jch Exp
 */

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/signalvar.h>
#include <sys/syslog.h>
#include <sys/select.h>
#include <sys/file.h>

#include <machine/cpu.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/netisr.h>
#include <net/route.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#endif

#ifdef NS
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif

#include <net/bpf.h>
#include <net/bpfdesc.h>

#include <net/if_tun.h>

struct tun_softc {
	struct	ifnet tun_if;		/* the interface */
	struct	ifqueue tun_fastq;	/* interactive output queue */
	u_int	tun_flags;		/* misc flags */
#define	TUN_OPEN	0x0001
#define	TUN_INITED	0x0002
#define	TUN_RCOLL	0x0004
#define	TUN_IASET	0x0008
#define	TUN_DSTADDR	0x0010
#ifdef notdef
#define	TUN_READY	0x0020
#else
#define TUN_READY       (TUN_IASET|TUN_OPEN|TUN_DSTADDR)
#endif
#define	TUN_RWAIT	0x0040
#define	TUN_ASYNC	0x0080
#define	TUN_NBIO	0x0100
	int	tun_pgrp;		/* the process group - if any */
	struct	selinfo	tun_rsel;	/* read select */
	struct	selinfo	tun_wsel;	/* write select (not used) */
};

#define TUNDEBUG	if (tundebug) printf
int	tundebug = 0;

static int ntun;		/* number of tunnel interfaces (set at boot) */
static struct tun_softc *tun_softc;

static int	tunoutput __P((struct ifnet *, struct mbuf *, struct sockaddr *, struct rtentry *));
static int	tunselect __P((dev_t dev, int, struct proc *));
static int	tunopen __P((dev_t, int, int, struct proc *));
static int	tunclose __P((dev_t, int, int, struct proc *));
static int	tunread __P((dev_t, struct uio *, int));
static int	tunwrite __P((dev_t, struct uio *, int));
static int	tunifioctl __P((struct ifnet *, u_long, caddr_t));
static int	tunioctl __P((dev_t, u_long, caddr_t, int, struct proc *));
static int	tuninit __P((int));

struct devsw tunsw = {
	NULL,
	tunopen, tunclose, tunread, tunwrite, tunioctl, tunselect, nommap,
	nostrat, nodump, nopsize, 0,
	nostop
};

void
tunattach(n)
	int n;
{
	struct ifnet *ifp;
	struct tun_softc *tp;
	int i;

	ntun = n;
	tun_softc = malloc(n * sizeof(*tun_softc), M_DEVBUF, M_WAIT);
	bzero(tun_softc, n * sizeof(*tun_softc));
	for (i = 0, tp = tun_softc; i < ntun; i++, tp++) {
		tp->tun_flags = TUN_INITED;
		ifp = &tp->tun_if;
		ifp->if_name = "tun";
		ifp->if_unit = i;
		ifp->if_mtu = TUNMTU;
		ifp->if_type = IFT_PROPVIRTUAL;
		ifp->if_flags = IFF_POINTOPOINT | IFF_MULTICAST;
		ifp->if_ioctl = tunifioctl;
		ifp->if_output = tunoutput;
		tp->tun_fastq.ifq_maxlen = 32;	/* XXX */
		if_attach(ifp);
		bpfattach(&ifp->if_bpf, ifp, DLT_NULL, sizeof(u_int));
	}
}

/*
 * tunnel open - must be superuser & the device must be
 * configured in
 */
static int
tunopen(dev, flag, mode, p)
	dev_t dev;
	int flag, mode;
	struct proc *p;
{
	struct ifnet *ifp;
	struct tun_softc *tp;
	int unit, error;

	if ((error = suser(p->p_ucred, &p->p_acflag)) != 0)
		return (error);

	if ((unit = minor(dev)) >= ntun)
		return (ENXIO);
	tp = &tun_softc[unit];
	if (tp->tun_flags & TUN_OPEN)
		return ENXIO;
	ifp = &tp->tun_if;
	tp->tun_flags |= TUN_OPEN;
	TUNDEBUG("%s%d: open\n", ifp->if_name, ifp->if_unit);
	return (0);
}

/*
 * tunclose - close the device - mark i/f down & delete
 * routing info
 */
static int
tunclose(dev, flag, mode, p)
	dev_t dev;
	int flag, mode;
	struct proc *p;
{
	int unit = minor(dev), s;
	struct tun_softc *tp = &tun_softc[unit];
	struct ifnet *ifp = &tp->tun_if;
	struct mbuf *m;

	tp->tun_flags &= ~TUN_OPEN;

	/*
	 * junk all pending output
	 */
	do {
		s = splimp();
		IF_DEQUEUE(&tp->tun_fastq, m);
		if (m == NULL)
			IF_DEQUEUE(&ifp->if_snd, m);
		splx(s);
		if (m != NULL)
			m_freem(m);
	} while (m != NULL);

	if (ifp->if_flags & IFF_UP) {
		s = splimp();
		if_down(ifp);
		splx(s);
	}
	tp->tun_pgrp = 0;
	selwakeup(&tp->tun_rsel);
		
	TUNDEBUG ("%s%d: closed\n", ifp->if_name, ifp->if_unit);
	return (0);
}

static int
tuninit(unit)
	int unit;
{
	struct tun_softc *tp = &tun_softc[unit];
	struct ifnet *ifp = &tp->tun_if;
	struct ifaddr *ifa;

	TUNDEBUG("%s%d: tuninit\n", ifp->if_name, ifp->if_unit);

	ifp->if_flags |= IFF_RUNNING;
	if_up(ifp);

	/* Verify that both a source and destination address are configured */
	for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
		if (ifa->ifa_addr != NULL &&
		    ifa->ifa_addr->sa_family != AF_LINK)
			tp->tun_flags |= TUN_IASET;
		if (ifa->ifa_dstaddr != NULL &&
		    ifa->ifa_dstaddr->sa_family != AF_LINK)
			tp->tun_flags |= TUN_DSTADDR;
	}

	return 0;
}

/*
 * Process an ioctl request.
 */
static int
tunifioctl(ifp, cmd, data)
	struct ifnet *ifp;
	u_long cmd;
	caddr_t	data;
{
	struct tun_softc *tp = &tun_softc[ifp->if_unit];
	struct ifreq *ifr;
	int error = 0, s;

	s = splimp();
	switch(cmd) {
	case SIOCSIFADDR:
		tuninit(ifp->if_unit);
		break;

	case SIOCSIFDSTADDR:
		tp->tun_flags |= TUN_DSTADDR;
		TUNDEBUG("%s%d: destination address set\n", ifp->if_name,
		    ifp->if_unit);
		break;

	case SIOCADDMULTI:
	case SIOCDELMULTI:
		ifr = (struct ifreq *)data;
		if (ifr == NULL) {
			error = EAFNOSUPPORT;		/* XXX */
			break;
		}
		switch (ifr->ifr_addr.sa_family) {

#ifdef INET
		case AF_INET:
			break;
#endif

		default:
			error = EAFNOSUPPORT;
			break;
		}
		break;

	default:
		error = EINVAL;
	}
	splx(s);
	return (error);
}

/*
 * tunoutput - queue packets from higher level ready to put out.
 */
static int
tunoutput(ifp, m0, dst, rt)
	struct ifnet *ifp;
	struct mbuf *m0;
	struct sockaddr *dst;
	struct rtentry *rt;
{
	struct tun_softc *tp = &tun_softc[ifp->if_unit];
	struct proc *p;
	struct ifqueue *ifq;
	int s;

	TUNDEBUG ("%s%d: tunoutput\n", ifp->if_name, ifp->if_unit);

	if ((tp->tun_flags & TUN_READY) != TUN_READY) {
		TUNDEBUG ("%s%d: not ready 0%o\n", ifp->if_name,
		    ifp->if_unit, tp->tun_flags);
		m_freem (m0);
		return EHOSTDOWN;
	}

	if (ifp->if_bpf != NULL) {
		/*
		 * We need to prepend the address family as
		 * a four byte field.  Cons up a dummy header
		 * to pacify bpf.  This is safe because bpf
		 * will only read from the mbuf (i.e., it won't
		 * try to free it or keep a pointer to it).
		 */
		struct mbuf m;
		u_int af = dst->sa_family;

		m.m_next = m0;
		m.m_len = sizeof(af);
		m.m_data = (char *)&af;

		bpf_mtap(ifp->if_bpf, &m);
	}

	ifq = &ifp->if_snd;
	switch(dst->sa_family) {
#ifdef INET
	case AF_INET:
		if (IPTOS_FASTQ(mtod(m0, struct ip *)))
			ifq = &tp->tun_fastq;
		s = splimp();
		if (IF_QFULL(ifq)) {
			IF_DROP(ifq);
			m_freem(m0);
			splx(s);
			ifp->if_collisions++;
			return (ENOBUFS);
		}
		IF_ENQUEUE(ifq, m0);
		splx(s);
		ifp->if_opackets++;
		break;
#endif
	default:
		m_freem(m0);
		return EAFNOSUPPORT;
	}

	if (tp->tun_flags & TUN_RWAIT) {
		tp->tun_flags &= ~TUN_RWAIT;
		wakeup(tp);
	}
	if (tp->tun_flags & TUN_ASYNC && tp->tun_pgrp) {
		if (tp->tun_pgrp > 0)
			gsignal(tp->tun_pgrp, SIGIO);
		else if ((p = pfind(-tp->tun_pgrp)) != 0)
			psignal(p, SIGIO);
	}
	selwakeup(&tp->tun_rsel);
	return 0;
}

/*
 * the cdevsw interface is now pretty minimal.
 */
static int
tunioctl(dev, cmd, data, flag, p)
	dev_t dev;
	u_long cmd;
	caddr_t data;
	int flag;
	struct proc *p;
{
	int unit = minor(dev), s;
	struct tun_softc *tp = &tun_softc[unit];
	struct tuninfo *tunp;

	switch (cmd) {
	case TUNSIFINFO:
	        tunp = (struct tuninfo *)data;
		tp->tun_if.if_mtu = tunp->mtu;
		tp->tun_if.if_type = tunp->type;
		tp->tun_if.if_baudrate = tunp->baudrate;
		break;
	case TUNGIFINFO:
		tunp = (struct tuninfo *)data;
		tunp->mtu = tp->tun_if.if_mtu;
		tunp->type = tp->tun_if.if_type;
		tunp->baudrate = tp->tun_if.if_baudrate;
		break;
	case TUNSDEBUG:
		tundebug = *(int *)data;
		break;
	case TUNGDEBUG:
		*(int *)data = tundebug;
		break;
	case FIONBIO:
		if (*(int *)data)
			tp->tun_flags |= TUN_NBIO;
		else
			tp->tun_flags &= ~TUN_NBIO;
		break;
	case FIOASYNC:
		if (*(int *)data)
			tp->tun_flags |= TUN_ASYNC;
		else
			tp->tun_flags &= ~TUN_ASYNC;
		break;
	case FIONREAD:
		s = splimp();
		if (tp->tun_fastq.ifq_head != NULL)
			*(int *)data = tp->tun_fastq.ifq_head->m_len;
		else if (tp->tun_if.if_snd.ifq_head != NULL)
			*(int *)data = tp->tun_if.if_snd.ifq_head->m_len;
		else	
			*(int *)data = 0;
		splx(s);
		break;
	case TIOCSPGRP:
		tp->tun_pgrp = *(int *)data;
		break;
	case TIOCGPGRP:
		*(int *)data = tp->tun_pgrp;
		break;
	default:
		return (ENOTTY);
	}
	return (0);
}

/*
 * The cdevsw read interface - reads a packet at a time, or at
 * least as much of a packet as can be read.
 */
static int
tunread(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	int unit = minor(dev);
	struct tun_softc *tp = &tun_softc[unit];
	struct ifnet *ifp = &tp->tun_if;
	struct mbuf *m, *m0;
	int error = 0, len, s;

	TUNDEBUG ("%s%d: read\n", ifp->if_name, ifp->if_unit);
	if ((tp->tun_flags & TUN_READY) != TUN_READY) {
		TUNDEBUG ("%s%d: not ready 0%o\n", ifp->if_name,
			  ifp->if_unit, tp->tun_flags);
		return EHOSTDOWN;
	}

	tp->tun_flags &= ~TUN_RWAIT;

	s = splimp();
	do {
		IF_DEQUEUE(&tp->tun_fastq, m0);
		if (m0 == NULL)
			IF_DEQUEUE(&ifp->if_snd, m0);
		if (m0 == NULL) {
			if (tp->tun_flags & TUN_NBIO) {
				splx(s);
				return EWOULDBLOCK;
			}
			tp->tun_flags |= TUN_RWAIT;
			tsleep((caddr_t)tp, PZERO + 1, "tunread", 0);
		}
	} while (m0 == NULL);
	splx(s);

	while (m0 && uio->uio_resid > 0 && error == 0) {
		len = min(uio->uio_resid, m0->m_len);
		if (len == 0)
			break;
		error = uiomove(mtod(m0, caddr_t), len, uio);
		MFREE(m0, m);
		m0 = m;
	}

	if (m0) {
		TUNDEBUG("Dropping mbuf\n");
		m_freem(m0);
	}
	return error;
}

/*
 * the cdevsw write interface - an atomic write is a packet - or else!
 */
static int
tunwrite(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	int unit = minor(dev);
	struct ifnet *ifp = &tun_softc[unit].tun_if;
	struct mbuf *m;
	int error, s;

	TUNDEBUG("%s%d: tunwrite\n", ifp->if_name, ifp->if_unit);

	if (uio->uio_resid < 0 || uio->uio_resid > (int)ifp->if_mtu) {
		TUNDEBUG("%s%d: len=%d!\n", ifp->if_name, ifp->if_unit,
		    uio->uio_resid);
		return EIO;
	}

	/* get a header mbuf */
	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == NULL)
		return ENOBUFS;

	if (uio->uio_resid > 0) {
		if (uio->uio_resid >= MINCLSIZE) {
			MCLGET(m, M_DONTWAIT);
			if ((m->m_flags & M_EXT) == 0) {
				m_freem(m);
				return ENOBUFS;
			}
		}
		m->m_len = uio->uio_resid;
		error = uiomove(mtod (m, caddr_t), m->m_len, uio);
		if (error) {
			m_freem(m);
			return error;
		}
	}

	m->m_pkthdr.len = uio->uio_resid;
	m->m_pkthdr.rcvif = ifp;

	if (ifp->if_bpf != NULL) {
		/*
		 * We need to prepend the address family as
		 * a four byte field.  Cons up a dummy header
		 * to pacify bpf.  This is safe because bpf
		 * will only read from the mbuf (i.e., it won't
		 * try to free it or keep a pointer to it).
		 */
		struct mbuf m0;
		u_int af = AF_INET;

		m0.m_next = m;
		m0.m_len = 4;
		m0.m_data = (char *)&af;

		bpf_mtap(ifp->if_bpf, &m0);
	}

	s = splimp();
	if (IF_QFULL (&ipintrq)) {
		IF_DROP(&ipintrq);
		splx(s);
		ifp->if_collisions++;
		m_freem(m);
		return ENOBUFS;
	}
	IF_ENQUEUE(&ipintrq, m);
	splx(s);
	ifp->if_ipackets++;
	schednetisr(NETISR_IP);
	return error;
}

/*
 * tunselect - the select interface, this is only useful on reads
 * really. The write detect always returns true, write never blocks
 * anyway, it either accepts the packet or drops it.
 */
static int
tunselect(dev, rw, p)
	dev_t dev;
	int rw;
	struct proc *p;
{
	int unit = minor(dev), s;
	struct tun_softc *tp = &tun_softc[unit];
	struct ifnet *ifp = &tp->tun_if;

	s = splimp();
	TUNDEBUG("%s%d: tunselect\n", ifp->if_name, ifp->if_unit);
	switch (rw) {
	case FREAD:
		if (tp->tun_fastq.ifq_len > 0 || ifp->if_snd.ifq_len > 0) {
			splx(s);
			TUNDEBUG("%s%d: tunselect fq=%d q=%d\n", ifp->if_name,
			    ifp->if_unit, tp->tun_fastq.ifq_len,
			    ifp->if_snd.ifq_len);
			return 1;
		}
		selrecord(curproc, &tp->tun_rsel);
		break;
	case FWRITE:
		splx(s);
		return 1;
	}
	splx(s);
	TUNDEBUG("%s%d: tunselect waiting\n", ifp->if_name, ifp->if_unit);
	return 0;
}
