/*
   Pseudo-device for Network Tracing (NTR)
   1. Add a pseudo-device name in /sys/conf/file
   pseudo-device ntr
   2. Add a pseudo-device structure in /sys/i386/conf/ioconf.c.i386
   %DEVSW(ntrsw)
   3. Define pseudo-device with all 6 basic functions: (see below)
   struct devsw ntrsw = { ... }
   4. Define an device attach function
   int ntrattach
   
   Specific to this pseudo-device:
   5. Register your service (open, close, read, write) in the ntr_servtab
*/

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/time.h>
#include <sys/proc.h>
#include <sys/user.h>
#include <sys/ioctl.h>
#include <sys/map.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/protosw.h>
#include <sys/errno.h>
#include <sys/select.h>

#include "ntr.h"
#include <net/ntrvar.h>


/* One descriptor per open device file. */
struct ntr_d {
	int busy;
	struct selinfo select;
};

struct ntr_serv {
	int (*open)	__P((dev_t, int, int, struct proc *));
	int (*close)	__P((dev_t, int, int, struct proc *));
	int (*read)	__P((dev_t, struct uio *, int));
	int (*write)	__P((dev_t, struct uio *, int));
	int (*ioctl)	__P((dev_t, u_long, caddr_t, int, struct proc *));
	int (*select)	__P((dev_t, int, struct proc *));
};


struct ntr_serv ntr_servtab[] = {
#if	NNTR > 0
	{ ntrace_open, ntrace_close, ntrace_read, NULL, NULL, NULL },
#endif
#if	NNTR > 1
	{ NULL, NULL, NULL, NULL, NULL, NULL },
#endif
};


int ntropen	__P((dev_t, int, int, struct proc *));
int ntrclose	__P((dev_t, int, int, struct proc *));
int ntrread	__P((dev_t, struct uio *, int));
int ntrwrite	__P((dev_t, struct uio *, int));
int ntrselect	__P((dev_t, int, struct proc *));
int ntrioctl	__P((dev_t, u_long, caddr_t, int, struct proc *));

#if	!defined(__NetBSD__)
struct devsw ntrsw = {
	NULL,
	ntropen, ntrclose, ntrread, ntrwrite, ntrioctl, ntrselect,
	nommap, nostrat, nodump, nopsize, 0, nostop
};
#endif


struct ntr_d *ntr_dtab;		/* table of minors for this device */
static int num_minor;		/* number of minors for this device */


/* Called at boot time, to establish the pseudo-device. */
int
ntrattach(int num)
{
	register struct ntr_d *d;
	register int i;

	ntr_dtab = malloc(num * sizeof(*ntr_dtab), M_DEVBUF, M_WAIT);
	num_minor = num;
	for (d=ntr_dtab, i=0;  i < num;  d++, i++) {
		d->busy = 0;
	}
	return 0;
}


int
ntropen(dev_t dev, int flag, int ifmt, struct proc *p)
{
	register struct ntr_d *d;
	int minor = minor(dev);
	int error = 0;

	if (minor >= num_minor)
		return ENXIO;

	/* Only 1 process per minor. If in use, return EBUSY */
	d = &ntr_dtab[minor];
	if (d->busy)
		return EBUSY;
	bzero(d, sizeof(*d));
	d->busy = 1;

	/* Initialize your SERVICE */
	if (ntr_servtab[minor].open)
		(*(ntr_servtab[minor].open))(dev, flag, ifmt, p);
	return error;
}


int
ntrclose(dev_t dev, int flag, int ifmt, struct proc *p)
{
	int s;
	int minor = minor(dev);
	struct ntr_d *d = &ntr_dtab[minor];

	/* End your SERVICE */
	if (ntr_servtab[minor].close)
		(*(ntr_servtab[minor].close))(dev, flag, ifmt, p);

	s = splimp();
	d->busy = 0;
	splx(s);
	return 0;
}


int
ntrread(dev_t dev, struct uio *uio, int ioflag)
{
	int status;
	int minor = minor(dev);

	/* Read request for your SERVICE */
	if (ntr_servtab[minor].read)
		status = (*(ntr_servtab[minor].read)) (dev, uio, ioflag);
	else return ENODEV;
	return status;
}


int
ntrwrite(dev_t dev, struct uio *uio, int ioflag)
{
	int status;
	int minor = minor(dev);

	/* Write request for your SERVICE */
	if (ntr_servtab[minor].write)
		status = (*(ntr_servtab[minor].write)) (dev, uio, ioflag);
	else return ENODEV;
	return status;
}


int
ntrioctl(dev_t dev, u_long cmd, caddr_t addr, int flag, struct proc *p)
{
	int error = 0;
	
	switch (cmd) {
	default:
		error = EINVAL;
		break;

	case FIONBIO:
	case FIOASYNC:
		break;

	}
}	


/* Support for select() system call. */
int
ntrselect(dev_t dev, int rw, struct proc *p)
{
	register struct ntr_d *d;
	register int s;

	if (rw != FREAD)
		return 0;

	d = &ntr_dtab[minor(dev)];

	s = splimp();
	/* No data ready. */
	selrecord(p, &d->select);
	splx(s);
	return 0;
}
