#ifndef _BLURB_
#define _BLURB_
/*
# 		    Ntrace: Mobile Network Tracing
# 			  Release 1.0-alpha
# 
# Copyright (c) 1996 Carnegie Mellon University, Regents of the
# University of California.  All Rights Reserved
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#  1. Redistributions of source code of this work, derivative works, or
#     modified versions must retain the above copyright notice, this
#     list of conditions and the following disclaimer. 
#  2. Redistributions in binary form of this work, derivative works or
#     modified versions must reproduce the above copyright notice, this
#     list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the
#     distribution.  
#  3. All advertising materials mentioning features or use of this
#     software must display the following acknowledgement:
# 	This product includes software developed by Carnegie Mellon
# 	University, the University of California, Berkeley, and their
# 	contributors.  
#  4. Neither the names of the universities nor the names of their
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
# 
#   NTRACE IS AN EXPERIMENTAL SOFTWARE SYSTEM AND IS KNOWN TO HAVE BUGS,
#   SOME OF WHICH MAY HAVE SERIOUS CONSEQUENCES.  CARNEGIE MELLON, THE
#   REGENTS, AND THEIR CONTRIBUTORS ALLOW FREE USE OF THIS SOFTWARE IN
#   ITS "AS IS" CONDITION.  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
#   BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
#   CARNEGIE MELLON, THE REGENTS, OR THEIR CONTRIBUTORS BE LIABLE FOR ANY
#   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
#   GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
#   IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE OR OF
#   ANY DERIVATIVE WORK, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
#   DAMAGE.
# 
# Carnegie Mellon and the Regents encourage users of this software to
# return any improvements or extensions that they make, and to grant
# Carnegie Mellon and the Regents the rights to redistribute these
# changes without encumbrance.
# 
*/
#endif /*_BLURB_*/







#ifndef _NTRREC_H_
#define _NTRREC_H_

#define NTR_VERSION   4

#include "ntrace.h"

/**********************************************************************
 * The basic idea: each trace is fronted by a trace header, and
 * contains one or more "packet TRACKs" and one or more "device
 * TRACKs."  Each packet TRACK is intended to encapsulate a single
 * protocol with specific measurement granularities.  Each device
 * TRACK is intended to encapsulate a single physical device (perhaps
 * a single <local-device, remote-device> pair) with specific
 * measurment granularities.  Each "TRACK" is begun with a "TRACK
 * header," which specifies the granularites of the observations and
 * types of protocol/device. 
 * 
 * I've followed Satya's advice here and made every number 32 bits --
 * we can haggle over bits later, and/or employ a simple compression
 * scheme to make things more space-efficient.
 *
 * Each of these 32-bit numbers should be stored on disk in
 * network order!!!
 *
 * Any record that is variable-sized must be padded to the nearest
 * word boundary.
 *
 * A note about timestamps.  All timestamps in a trace follow the same
 * format. Two words; the first is always seconds past the UNIX Zero
 * Hour, the second is in either usecs or nsecs past that number of
 * seconds.  All timestamps in a trace use *either* usec fractions, or
 * nsec fractions.  If two traces are merged, the result will be in usecs
 * if *either* of the sources were in usecs
 ************************************************************************/

/************************************************************************
 * Magic numbers 
 *
 * All trace records (and some components of same) are identified by
 * pseudo-random magic numbers.  These are stored on-disk in network
 * order to preserve portability.  They are #define'd rather than
 * enum'd to ensure that the sizes of the magic numbers are known and
 * portable.
 * 
 *************************************************************************/

/* Trace header record */
#define TR_HEADER_MAGIC    0xC2B1382B
/* End Mark */
#define TR_TRAILER_MAGIC   0xC2B8B772

/* TRACK header records */
#define TR_PKT_HDR_MAGIC   0xC2B7A5D5
#define TR_DEV_HDR_MAGIC   0xC2B6E8A2
#define TR_GEN_HDR_MAGIC   0xC2BC19FE

/* Annotations */
#define TR_ANNOTE_MAGIC    0xC2BFF7A0

/* Loss records */
#define TR_LOSS_MAGIC      0xC2BA4A24

/* Protocol types: which protocols we are capable of tracing */

#define TR_PKT_IP_MAGIC    0xC2D003BF	/* IP */
#define TR_PKT_TCP_MAGIC   0xC2D0B19C	/* TCP */
#define TR_PKT_UDP_MAGIC   0xC2D0413D	/* UDP */
#define TR_PKT_ICMP_MAGIC  0xC2D004E1	/* ICMP */

/* Device types: which devices we know about at the moment. */

#define TR_DEV_EE_MAGIC    0xC2D5A3D2	/* Intel EtherExpress */
#define TR_DEV_EP_MAGIC    0xC2D514F9	/* 3Com EtherLink III */
#define TR_DEV_WL_MAGIC    0xC2D52C3C	/* NCR WaveLan */
#define TR_DEV_WG_MAGIC    0xC2D56315	/* NCR WaveLan 2.4 GHz*/
#define TR_DEV_MC_MAGIC    0xC2D55B19	/* MetriCom */
#define TR_DEV_CD_MAGIC    0xC2D5B1FB	/* CDPD */
#define TR_DEV_SL_MAGIC    0xC2D5FBFA	/* Slip -- Modem */

/******************************************************************
 * Property Lists
 *
 * Some data in packets is required, and other fields are optional.
 * To provide self-describing traces, TRACK headers contain "property
 * lists" which specify optional data.  Some of these optional data
 * fields are "header-only," in that they are only included in the
 * TRACK header, others are "entry" properties; they are included in
 * each entry in this TRACK.
 *
 * Property lists have a flags field, a type field specifying the type
 * in which the observation is expressed, the size of each observation
 * in bytes, the name of the observation recorded, and possibly the
 * value itself.  Header-only properties have a flag set in their
 * flags field; in such cases, the property list contains the
 * observation itself.  Otherwise, the value is recorded in each track
 * entry record.
 *
 * There is a structure in the libntrrec.a which has names for the
 * common properties contained here.  If a property isn't present in 
 * this file, it still may appear in a trace, but has to follow the
 * format rules here.
 *********************************************************************/

/* 
 * Defined flags 
 */
#define PLIST_HEADER_ONLY   0x1

/* 
 * Defined types
 */
#define PLIST_TYPE_UINT8    0x1
#define PLIST_TYPE_INT8     0x2
#define PLIST_TYPE_UINT16   0x3
#define PLIST_TYPE_INT16    0x4
#define PLIST_TYPE_UINT32   0x5
#define PLIST_TYPE_INT32    0x6
#define PLIST_TYPE_UINT64   0x7
#define PLIST_TYPE_INT64    0x8
#define PLIST_TYPE_STRING   0x9

#define PAD_BYTES(sz)   (((sz) % sizeof(u_int32_t)) \
			 ? (sizeof(u_int32_t) - ((sz) % sizeof(u_int32_t))) \
			 : 0)
			 

struct tr_plist_ent_t {
    u_int8_t     pl_flags;   /* header-only? */
    u_int8_t     pl_type;    /* Type of recorded data */
    u_int16_t    pl_size;    /* Size of data in bytes */
    u_int32_t    pl_name;    /* Which option recorded */
    u_int8_t     *pl_value;  /* Only if header prop. */
};


/* 
 * Options -- things named in property lists -- that are currently defined
 */

#define ADDR_SEND       0xA054B078   /* IP address of sending host */
#define ADDR_RECV       0xA054976E   /* IP address of recieving host */
#define ADDR_LINK       0xA054E821   /* IP address of first/last-hop router */
#define PKT_SEQUENCE    0xA0549C30   /* Sequence number of packet */
#define PKT_SENTTIME    0xA0540267   /* Time packet was sent */
#define PKT_HOPS        0xA0542F7E   /* Number of hops packet took */
#define SOCK_SENDPORT   0xA05511D9   /* Sending socket's port */
#define SOCK_RECVPORT   0xA0551125   /* Recieving socket's port */
#define PROTO_FLAGS     0xA03055E0   /* Protocol-specific flags */
#define PROTO_ERRLIST   0xA03089C1   /* Protocol-specific errors */
#define IP_PROTO        0xA0315CA6   /* IP protocol number */
#define ICMP_PINGTIME   0xA034F638   /* ICMP ECHO roundtrip time */
#define ICMP_KIND       0xA034E9D3   /* ICMP 'kind' field */
#define ICMP_ID         0xA0347F10   /* ICMP 'id' field */
#define TCP_MSSIZE      0xA032CFED   /* TCP's maximum segment size */
#define TCP_DPORT       0xA032651E   /* TCP's destination port */
#define TCP_SPORT       0xA0328F8A   /* TCP's sending port */
#define TCP_ACK         0xA032B064   /* TCP's ack number in this packet */
#define TCP_WINDOW      0xA032A7CE   /* Current window size */
#define TCP_TIMESTAMP   0xA032371C   /* Timestamp in TCP packet */
#define DEV_MAJID       0xA0609E91   /* Major id of device */
#define DEV_MINID       0xA0601B32   /* Minor id of device */
#define DEV_STATUS      0xA0600A79   /* Device-specific status flags */
#define WL_SIG_LEVEL	0xA0612D1E   /* WaveLAN signal level */
#define WL_SIG_QUAL	0xA06108CD   /* WaveLAN signal quality measure */
#define WL_SIL_LEVEL	0xA0610EE3   /* WaveLAN silence level */
#define MC_RSSI		0xA0623364   /* Metricom signal-to-noise ratio */


/* Error types */
#define TR_EDEV_UNKNOWN		0x10
#define TR_EDEV_BUF		0x11
#define TR_EDEV_FRAME		0x12
#define TR_EDEV_CRC		0x14
#define TR_EPKT_IP_HDR		0x40
#define TR_EPKT_IP_HDR_CRC	0x41
#define TR_EPKT_IP_HDR_LEN	0x42
#define TR_EPKT_IP_HDR_VER	0x44
#define TR_EPKT_TCP_HDR 	0x80
#define TR_EPKT_TCP_DATA_CRC	0xc0

/* Specific macros for ICMP processing */
#define TR_ICMP_TYPE(kind)     (((kind) & 0xff00) >> 8)
#define TR_ICMP_CODE(kind)     ((kind) & 0xff)
#define TR_ICMP_INBOUND(kind)  ((int32_t)(kind) < 0) /* msb set for inbound */
#define TR_ICMP_OUTBOUND(kind) (!ICMP_INBOUND(kind)) /* ..clear for outbound */




/***********************************************************************
 * Trace Entry Definitions
 **********************************************************************/

/* Major structures */

struct tr_header_t;		/* The full trace header */
struct tr_trailer_t;             /* The full trace footer */
struct tr_pkt_hdr_t;		/* Header for a new 'packet TRACK' */ 
struct tr_pkt_ent_t;		/* Packet trace entry */
struct tr_dev_hdr_t;		/* Header for a new 'device TRACK' */
struct tr_dev_ent_t;		/* Device trace entry */
struct tr_gen_hdr_t;		/* Header for a generic TRACK */
struct tr_gen_ent_t;		/* Trace entry in a generic TRACK */
struct tr_loss_t;		/* Lost data entry */
struct tr_annot_t;		/* Annotation */

/* An enum to define all known trace entry types */

enum tr_entry_types_t {
    tr_ent_unknown=0,
    tr_ent_header,
    tr_ent_trailer,
    tr_ent_pkthdr,
    tr_ent_devhdr,
    tr_ent_genhdr,
    tr_ent_pktent,
    tr_ent_devent,
    tr_ent_genent,
    tr_ent_annote,
    tr_ent_loss
};

/* 
 * An anonymous trace record -- all trace records start with three
 * words: magic and size -- makes for easy sanity checking, etc.  Can
 * always cast a trace record to this to see what it is.  
 */

struct tr_unknown_t {
    u_int32_t	     un_magic;
    u_int32_t	     un_size;
    struct tr_time_t un_timestamp;
};


/* 
 * The biggest we'll ever let a trace record get.  It is intentionally
 * generous, and only used as a limit to how big various buffers
 * etc. can grow.  It allows room for a few hundred thousand
 * property-list entries
 */
#define TR_MAX_RECORD_SIZE  0x100000

#define TR_SIZE(p)	(((struct tr_unknown *) (p))-> un_size)
#define TR_SIZEOF(t)	(sizeof(t) - sizeof(char*))

#define TR_SIZE_BAD(p)  (TR_SIZE(p) < TR_SIZEOF(*p) \
			 || (TR_SIZE(p) % sizeof(u_int32_t) != 0) \
			 || (TR_SIZE(p) > TR_MAX_RECORD_SIZE))

/*
 * tr_header_t contains some string fields; this is mostly to ease the
 * burden of identifying traces: we can use standard tools to discover
 * simple info about any trace.
 */
struct tr_header_t {
    u_int32_t          h_magic;
    u_int32_t          h_size;
    struct tr_time_t   h_timestamp;  /* starting time */
    u_int32_t          h_version;    /* Version # of trace */
    int16_t            h_minwest;    /* Minutes west of GMT */
    u_int16_t          h_dsttype;    /* Type of daylt. sav.*/
    struct tr_string_t h_date;       /* Date collected */
    struct tr_string_t h_desc;       /* description of trace */
    u_int32_t          h_numagent;   /* # hosts traced */
    struct tr_string_t *h_agent;     /* DNS name [numagent] */
    u_int32_t          *h_agent_ip;  /* IP addr [numagent] */
};


/* Minimum size: there might be no agents named in the trace. */

#define TR_HEADER_MINSZ     (sizeof(struct tr_header_t) \
			     - sizeof(struct tr_string_t *) \
			     - sizeof(u_int32_t *))
			    

/* It's bad if it's too small, too large, or not a multiple of a longword */

#define TR_HEADER_SIZE_BAD(hp)  (((hp)->h_size < TR_HEADER_MINSZ) \
				 || ((hp)->h_size % sizeof(u_int32_t) != 0)\
				 || ((hp)->h_size > TR_MAX_RECORD_SIZE))


/* Packet headers/entries */

struct tr_pkt_hdr_t {
    u_int32_t             ph_magic;
    u_int32_t             ph_size;
    struct tr_time_t      ph_timestamp;
    u_int32_t             ph_trackmagic; /* magic # of track ent. */
    u_int32_t             ph_ipaddr;     /* host generating track */
    u_int32_t             ph_dev_type;   /* device collected from */
    u_int32_t             ph_protocol;   /* protocol */
    u_int32_t             ph_plistsz;    /* Entries in plist */
    struct tr_plist_ent_t *ph_plist;     /* variable size */
};

/* Minimum size: no properties. */

#define TR_PKT_HDR_MINSZ  (sizeof(struct tr_pkt_hdr_t) \
			   - sizeof(struct tr_plist_ent_t *))

#define TR_PKT_HDR_SIZE_BAD(php) (((php)->ph_size < TR_PKT_HDR_MINSZ) \
				  || ((php)->ph_size % sizeof(u_int32_t) != 0)\
				  || ((php)->ph_size > TR_MAX_RECORD_SIZE))


#define TR_PKT_INBOUND    0x1
#define TR_PKT_OUTBOUND   0x2

struct tr_pkt_ent_t {
    u_int32_t        pe_magic;
    u_int32_t        pe_size;
    struct tr_time_t pe_timestamp;
    u_int32_t        pe_flags;
    u_int32_t        pe_psize;    /* packet size */
    u_int8_t         *pe_vlist;   /* determined by plist/size */
};

/* Minimum size: no values */
#define TR_PKT_ENT_MINSZ  (sizeof(struct tr_pkt_ent_t) \
			   - sizeof(u_int8_t *))

#define TR_PKT_ENT_SIZE_BAD(pep) (((pep)->pe_size < TR_PKT_ENT_MINSZ) \
				  || ((pep)->pe_size % sizeof(u_int32_t) != 0)\
				  || ((pep)->pe_size > TR_MAX_RECORD_SIZE))


/* Device headers/entries */

struct tr_dev_hdr_t {
    u_int32_t           dh_magic;
    u_int32_t           dh_size;
    struct tr_time_t    dh_timestamp;
    u_int32_t           dh_trackmagic; /* Magic # of track ent. */
    u_int32_t           dh_ipaddr;     /* host generating track */
    u_int32_t           dh_dev_type;   /* device described */
    u_int32_t           dh_plistsz;    /* Entries in plist */
    struct tr_plist_ent_t *dh_plist;   /* Variable size */
};

/* Minimum size: no properties. */

#define TR_DEV_HDR_MINSZ  (sizeof(struct tr_dev_hdr_t) \
			   - sizeof(struct tr_plist_ent_t *))

#define TR_DEV_HDR_SIZE_BAD(dhp) (((dhp)->dh_size < TR_DEV_HDR_MINSZ) \
				  || ((dhp)->dh_size % sizeof(u_int32_t) != 0)\
				  || ((dhp)->dh_size > TR_MAX_RECORD_SIZE))
struct tr_dev_ent_t {
    u_int32_t        de_magic;
    u_int32_t        de_size;
    struct tr_time_t de_timestamp;
    u_int8_t         *de_vlist;  /* determined by plist/size */
};

/* Minimum size: no values */
#define TR_DEV_ENT_MINSZ  (sizeof(struct tr_dev_ent_t) \
			   - sizeof(u_int8_t *))

#define TR_DEV_ENT_SIZE_BAD(dep) (((dep)->de_size < TR_DEV_ENT_MINSZ) \
				  || ((dep)->de_size % sizeof(u_int32_t) != 0)\
				  || ((dep)->de_size > TR_MAX_RECORD_SIZE))

struct tr_gen_hdr_t {
    u_int32_t             gh_magic;
    u_int32_t             gh_size;
    struct tr_time_t      gh_timestamp;
    u_int32_t             gh_trackmagic;
    u_int32_t             gh_ipaddr;
    u_int32_t             gh_plistsz;
    struct tr_plist_ent_t *gh_plist; /* Variable size */
};

#define TR_GEN_HDR_MINSZ  (sizeof(struct tr_gen_hdr_t) \
			   - sizeof(struct tr_plist_ent_t *))

#define TR_GEN_HDR_SIZE_BAD(ghp) (((ghp)->gh_size < TR_GEN_HDR_MINSZ) \
				  || ((ghp)->gh_size % sizeof(u_int32_t) != 0)\
				  || ((ghp)->gh_size > TR_MAX_RECORD_SIZE))
struct tr_gen_ent_t {
    u_int32_t        ge_magic;
    u_int32_t        ge_size;
    struct tr_time_t ge_timestamp;
    u_int8_t         *ge_vlist; /* Variable size */
};

/* Minimum size: no values */
#define TR_GEN_ENT_MINSZ  (sizeof(struct tr_gen_ent_t) \
			   - sizeof(u_int8_t *))

#define TR_GEN_ENT_SIZE_BAD(gep) (((gep)->ge_size < TR_GEN_ENT_MINSZ) \
				  || ((gep)->ge_size % sizeof(u_int32_t) != 0)\
				  || ((gep)->ge_size > TR_MAX_RECORD_SIZE))

/* Esoterics -- annotations and losses */

struct tr_annote_t {
    u_int32_t        a_magic;
    u_int32_t        a_size;
    struct tr_time_t a_timestamp;
    u_int32_t        a_ipaddr;
    struct tr_plist_ent_t a_plent; /* single, hdr plist ent. */
};

#define TR_ANNOTE_MINSZ   (sizeof(struct tr_annote_t) \
			   - sizeof(u_int8_t *))

#define TR_ANNOTE_SIZE_BAD(ap)  (((ap)->a_size < TR_ANNOTE_MINSZ)  \
				 || ((ap)->a_size % sizeof(u_int32_t) != 0) \
				 || ((ap)->a_size > TR_MAX_RECORD_SIZE))


struct tr_lossent_t {
    u_int32_t   le_magic; /* Magic # of records lost */
    int32_t     le_lost;  /* how many (-1 == don't know) */
};

struct tr_loss_t {
    u_int32_t        l_magic;
    u_int32_t        l_size;
    struct tr_time_t l_timestamp;
    u_int32_t        l_ipaddr;
    u_int32_t        l_lossentsz;  /* How many lossents */
    struct tr_lossent_t *l_lossarr; /* Array of lossents */
};

/* Minimum size is one lossent */
#define TR_LOSS_MINSZ  (sizeof(struct tr_loss_t) \
			- sizeof(struct tr_lossent_t *) \
			+ sizeof(struct tr_lossent_t)) 

#define TR_LOSS_SIZE_BAD(lp)   (((lp)->l_size < TR_LOSS_MINSZ) \
				|| ((lp)->l_size > TR_MAX_RECORD_SIZE) \
				|| (((lp)->l_size - TR_LOSS_MINSZ) \
				    % sizeof(struct tr_lossent_t) != 0))


struct tr_trailer_t {
    u_int32_t          t_magic;
    u_int32_t          t_size;
    struct tr_time_t   t_timestamp; /* end time */
    struct tr_string_t t_date;      /* Date end written */
};

#define TR_TRAILER_MINSZ   (sizeof(struct tr_trailer_t) - sizeof(char *))

#define TR_TRAILER_SIZE_BAD(tp)  (((tp)->t_size < TR_TRAILER_MINSZ) \
				  || ((tp)->t_size % sizeof(u_int32_t) != 0) \
				  || ((tp)->t_size > TR_MAX_RECORD_SIZE))

#endif /* _NTRREC_H_ */
