/*
 * Copyright (c) 1982, 1986, 1988, 1990, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Venkata N. Padmanabhan (padmanab@cs.berkeley.edu)
 */

/*
 * This is a bare-bones implementation of the Indirect-TCP (I-TCP)
 * protocol from Rutgers University. Essentially, this splices together
 * two TCP connections without involving any extra data copies or
 * user-kernel boundary crossings. A typical use of such a mechanism
 * is to splice together the wired and wireless portions of a connection
 * at a base station. This implementation was used for experiments in
 * a comparitive study of various TCP-over-wireless solutions. The
 * SIGCOMM '96 paper that resulted from this work is available at the
 * Web page listed below.
 *
 * For questions/comments about this implementation
 * Venkata N. Padmanabhan (padmanab@cs.berkeley.edu)
 * http://www.cs.berkeley.edu/~padmanab
 */ 

#ifdef ITCPEMU

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/errno.h>

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>
#include <netinet/tcp_debug.h>


/* the sockets for the wired and the wireless connections in I-TCP */
struct socket *itcp_wired_socket=NULL;
struct socket *itcp_wless_socket=NULL;

int itcp_wired_conn_closed=0; /* indicates if the wired connection is closed */

/*
 * Transfer as much data as possible from the receive buffer of the wired
 * connection to the send buffer of the wireless connection. If a non-zero
 * amount of data is transferred, invoke PRU_RCVD processing on wired socket
 * and PRU_SEND processing on the wireless socket.
 */
void
itcp_pipe_data() 
{
     struct socket *so1, *so2;
     struct mbuf   *m, *mm, *nextrecord;
     struct mbuf   **mp = &mm;
     struct mbuf   **mp0 = &mm;
     int count=0, len, so2_space;

     so1 = itcp_wired_socket;   
     so2 = itcp_wless_socket;

     if (!so1 || !so2)
	     return;

     /*
      * Loop thru the mbufs in the receive buffer of so1. Keep accumulating
      * the mbufs (and possibly a portion of the mbuf for the last piece)
      * into the chain pointed to by *mp0, so long as we do not exceed the 
      * space available in the send buffer of so2. The variable 'count' keeps
      * track of the number of bytes transferred.
      *
      * (code adapted from from kern/uipc_socket.c:soreceive() -- Stevens
      * Vol. 2, pp. 520-521)
      *
      * NOTE: We ignore OOB data.
      */

     so2_space = sbspace(&so2->so_snd);   /* free space available */  

     m = so1->so_rcv.sb_mb;
     while (m && (count < so2_space)) {
	  len = so2_space - count;
	  if (len > m->m_len)
	       len = m->m_len;
	  count += len;

	  if (len == m->m_len) { /* the entire mbuf is being transferred */
	       nextrecord = m->m_nextpkt;
	       sbfree(&so1->so_rcv, m);
	       /* add the mbuf m to the chain pointed to by *mp */
	       *mp = m;
	       mp = &m->m_next;
	       so1->so_rcv.sb_mb = m = m->m_next;
	       *mp = (struct mbuf *)0;
	       if (m)
		    m->m_nextpkt = nextrecord;
	  } else { /* try to make a copy of the partial mbuf to be transferred */ 
	       if (*mp = m_copym(m, 0, len, M_DONTWAIT)) { /* success */
		    m->m_data += len;
		    m->m_len -= len;
		    so1->so_rcv.sb_cc -= len;
	       }
	       else {  /* failure -- unable to create a new mbuf */
		    count -= len; 
	       }
	  }
     }
     /* 
      * If a non-zero amount of data has been accumulated in the chain pointed
      * to by *mp0, invoke PRU_RCVD on so1 (so that a window update can be 
      * sent to the original sender) and invoke PRU_SEND on so2 (to transfer
      * the data to the send buffer and send it to the eventual recipient).
      * If a FIN or RST has been received on the wired connection, close the
      * wireless connection as well.
      */
     if (count) {
	  tcp_usrreq(so1, PRU_RCVD, NULL, NULL, NULL);
	  tcp_usrreq(so2, PRU_SEND, *mp0, NULL, NULL);
	  /* 
	   * If wired connection has been closed and the corresponding
	   * buffer is empty, close the wireless connection.
	   */
	  if ((itcp_wired_conn_closed) && (so1->so_rcv.sb_cc == 0)) {
	       tcp_usrreq(so2, PRU_DISCONNECT, NULL, NULL, NULL); 
	       itcp_wired_conn_closed = 0;
	  }
	  
     }
}

#endif /* ITCPEMU */
