#ifndef _NETINET_STP_H_
#define _NETINET_STP_H_

/***************************************************************/ 
/*                                                             */ 
/* 	stp.h:  Configurable parameters, constants, macros,    */ 
/*			function prototypes                    */
/*                                                             */ 
/***************************************************************/ 

/*
 * General configuration debugging parameters (define STP_STATS, STP_DEBUG)
 * Define STP_DEBUG to enable debugging/error printfs
 * Define STP_STATS to enable detailed connection statistics (on stderr) 
 * Define STP_STATS_SUMMARY for a reduced version of the statistics
 * Define STP_THRESHOLD to be something other than 65536 in order to enable
 *     simulated random packet losses.  STP_THRESHOLD must be defined.
 */
/* #define STP_DEBUG */
/* #define STP_STATS */
/* #define STP_STATS_SUMMARY */
#define STP_THRESHOLD 65536 	/* 
				 * Determines incoming packet drop rate. 
				 * STP_THRESHOLD/65536 determines percentage
				 * of packets successfully delivered, e.g.,
			 	 * 65536 = 100%, 64881 = 99%, 58982 = 90%, etc.
				 */

/*
 * Implementation options
 */
#define WINDOW_CONTROL 1	/* Use TCP-like congestion control */
#define RATE_CONTROL 0		/* Bound upper and lower sending rate */
#define RATE 20000000		/* Bits per second */
#define WAIT_FOR_DATA 1 	/* Wait for process to send data before bgnak */
#define WAIT_FOR_DATA_TO 100 	/* (ms) wait period for process to send data */
#define CLOSE_ON_QUENCH 1	/* If TRUE, shuts cwnd to 1 upon source quench*/
#define ERROR_DETECTION 0	/* (NOT YET DEFINED) Receiver indicates to */
				/* sender if a missing PDU was likely  */
				/* errored, not lost */

/* 
 * Configuration constants 
 */
#define INITIAL_WINDOW 10	/* Send window used before BGN/BGNAK received */
#define INITIAL_SRTT 1000	/* 1000 ms */
#define INITIAL_CWND 1		/* Number of segments cwnd initialized to */
#define MIN_POLL_INTERVAL 50  	/* Minimum polling interval (ms) for short rtt*/
#define MAXBUFFER 1024
#define NUM_DUP_ACKS 3
#define SDP_THRESHOLD 10   	/* Piggyback POLL rather than regular POLL if */
				/* under this threshold */
#define MAXBURST 8		/* Maximum burst size in a 10ms interval */
#define MAXSTAT 7 		/* Max. number of list elements in a STAT; */
				/* Must be odd, 9 to fit in mbuf */
#define MAXBACKOFF 10		/* Maximum backoff is 2^MAXBACKOFF */ 
#define STP_MSS 1428 		/* 1428 + 16 (SDP) + 20 for IP = 1464 */
#define STP_PERSIST 120		/* 60 seconds between persist probes */ 
#define STP_KEEPALIVE 14400	/* 2 hours between keepalive probes */ 
#define STP_KEEPALIVE_INT 150	/* 75 seconds waiting for keepalive response */
#define STP_KEEPALIVE_MAX 12	/* Try 12 times in keepalive before giving up*/ 
#define STP_MAXRETRANS 12	/* Maximum number of retransmissions before drop */
#define STP_REROUTE 4		/* Call in_losing() upon this many retrans. */
#define STP_LINGERTIME 12000 	/* 2 minutes */
#define MAX_CC_ATTEMPTS 6	/* backoffs of 2,4,8,16,32= 63 s. before */ 
				/* quitting */
#define STP_MAXWIN 65535	/* Maximum window size (packets) */

/* 
 * Constants and (relatively) static parameters 
 */
#define TRUE 1
#define FALSE 0
#define STAT 0
#define USTAT 1
#define NO_POLL 0 
#define POLL 1
#define SRTT_SHIFT 3		/* srtt resolution = (1 ms >> SRTT_SHIFT) */ 
#define MODULO 16777216		/* 2^24 */
#define MS_PER_TICK 10		/* 10 ms per clock tick */
#define STP_NODELAY 0x01    	/* don't delay send to coalesce packets */
#define STP_MAXSEG 0x02 	/* set max. segment size (setsocketopt) */

/*  
 * STP states
 */
#define STP_CLOSED 0x0001U    		/* 1: no connection state */
#define STP_LISTEN 0x0002U    		/* 2: recv listen, send nothing */ 
#define STP_BGN_SENT 0x0004U 		/* 4: normal active open */
#define STP_BGN_DATA_SENT 0x0008U 	/* 8: active accelerated open */
#define STP_BGN_DATA_END_SENT 0x0010U 	/* 16: active accelerated open/close */
#define STP_WAIT_FOR_DATA 0x0020U   	/* 32: wait for process to send data */
#define STP_WAIT_FOR_DATA_CLOSE 0x0040U	/* 64: wait for process; move to */
					/* CLOSE_RECV afterwards (END recvd.)*/ 
#define STP_EST 0x0080U  		/* 128: Bidirectional established */
#define STP_CLOSE_RECV 0x0100U    	/* 256: read side closed by peer(END)*/
#define STP_HALF_CLOSE_PENDING 0x0200U	/* 512: write side closed by user */
#define STP_HALF_CLOSED 0x400U		/* 1024: write side closed by user */
#define STP_FULL_CLOSE_PENDING 0x0800U	/* 2048: both sides closed by user */
#define STP_FULL_CLOSED 0x1000U		/* 4096: wait for peer END*/
#define STP_LAST_ACK 0x2000U		/* 8192: waiting for ENDAK */

/*  
 * STP timers
 */ 
/* The following four timers are handled by the 10ms timeout routines */
#define STP_TIMER_SEND 0x01
#define STP_TIMER_BGN  0x02
#define STP_TIMER_END  0x04 
#define STP_TIMER_WAIT_BGNAK 0x08
/* The following two timers are handled by stp_slowtimo() */
#define STP_TIMER_PERSIST   0x10 
#define STP_TIMER_KEEPALIVE 0x20 
/*
 * out of band data
 */
#define STPOOB_HAVEDATA 0x01
#define STPOOB_HADDATA 0x02

/*  
 * STP PDU types (8 bit unsigned field; upper 3 bits presently unused)
 */ 
#define STP_PDU_SD 0x01U     
#define STP_PDU_POLL 0x02U
#define STP_PDU_SDP 0x03U 
#define STP_PDU_STAT 0x04U   
#define STP_PDU_USTAT 0x06U 
#define STP_PDU_POLL_URG 0x0AU
#define STP_PDU_SDP_URG 0x0BU

#define STP_PDU_BGN 0x10U
#define STP_PDU_BGN_SD 0x11U
#define STP_PDU_BGNAK 0x12U
#define STP_PDU_BGNAK_SD 0x13U
#define STP_PDU_BGN_SD_END 0x36U
#define STP_PDU_BGNAK_SD_END 0x37U
#define STP_PDU_END 0x24U
#define STP_PDU_SD_END 0x25U
#define STP_PDU_ENDAK 0x26U 
#define STP_PDU_ENDAK_SD 0x27U
#define STP_PDU_ENDAK_SD_END 0x29U

#define STP_PDU_RST 0x3FU


/* 
 * Macros 
 */
#define REASS_MBUF_STP(si) (*(struct mbuf **)&((si)->si_s))
#define	intostpcb(ip)	((struct stpcb *)(ip)->inp_ppcb)
#define	sotostpcb(so)	(intostpcb(sotoinpcb(so)))

#define SEQ_INC(a) (((a) + 1)%MODULO)
#define SEQ_DEC(a) (((a) - 1)%MODULO)

/* 
 * We want to avoid doing m_pullup on incoming packets but that
 * means avoiding dtom on the tcp reassembly code.  That in turn means
 * keeping an mbuf pointer in the reassembly queue (since we might
 * have a cluster).  As a quick hack, the source & destination
 * port numbers (which are no longer needed once we've located the
 * tcpcb) are overlayed with an mbuf pointer.
 */
#define STP_REASS_MBUF(si) (*(struct mbuf **)&((si)->si_s))

#ifdef STP_DEBUG
    void _Assert(char *, unsigned);
    #define ASSERT(f)   \
        if (f)          \
            {}          \
        else            \
            _Assert(__FILE__,__LINE__)
#else
    #define ASSERT(f)
#endif

#endif /* !_NETINET_STP_H_ */
