/*
 * Copyright (c) 1982, 1986, 1988, 1990, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Venkata N. Padmanabhan (padmanab@cs.berkeley.edu)
 */

/* 
 * This file contains functions that collect useful statistics pertaining
 * to a TCP connection. To the extent possible data is stored in memory, and
 * is written to a file only when either the in-memory data structure gets
 * filled up or when tracing is turned off for this connection (typically,
 * just before it is closed.
 *
 * For questions/comments, contact:
 * Venkata N. Padmanabhan (padmanab@cs.berkeley.edu)
 * http://www.cs.berkeley.edu/~padmanab
 */ 

#ifndef TUBA_INCLUDE
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/errno.h>

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>
#include <netinet/tcp_debug.h>

#include <netinet/tcp_stats.h>

#endif

extern       struct timeval time;

static struct timeval tcp_stats_starttime;    /* time when we start collecting
						 statistics */
/* array to store TCP-related statistics in memory */
static long         tcp_stats_buffer[TCP_STATS_BUFFER_SIZE]; 
static unsigned int tcp_stats_count;   /* # entries used up in array */
u_long              tcp_stats_flags;   /* indicates the fields of interest */

/*
 * Initialize the variables and data structures for collecting TCP
 * statistics.
 */
void tcp_stats_init(register struct tcpcb *tp, u_long flags) {
	tcp_stats_starttime = time;
	tcp_stats_count = 0;
	tcp_stats_flags = flags;
	tcp_stats_record(TCP_STATS_ISS|TCP_STATS_IRS, tp);
}

/*
 * Record relevant information about the current TCP event (as determined
 * by 'flags') in an in-memory data structure.
 */
void tcp_stats_record(unsigned long flags, register struct tcpcb *tp) {
	
	unsigned long localflags = flags & tcp_stats_flags;
	unsigned long onbit; /* position of the least significant ON bit */
	struct timeval curtime;
	
	/* 
	 * If we are close to filling up the in-memory array, dump to a file.
	 */
	if (tcp_stats_count >= TCP_STATS_BUFFER_THRESH) {
		tcp_stats_dump();
                /* XXX: remove return after a real dump function is written */
		return; 
	}
	
	/* 
	 * Record the flags for this entry.
	 */
	tcp_stats_buffer[tcp_stats_count++] = (long) localflags;
	microtime(&curtime);	/* record the current time */
	tcp_stats_buffer[tcp_stats_count++] = curtime.tv_sec;
	tcp_stats_buffer[tcp_stats_count++] = curtime.tv_usec;
	
	/* 
	 * Now we run thru all the bits in 'localflags' that are turned on 
	 * starting with the lsb. For each bit, record the corresponding
	 * field in tcp_stats_buffer.
	 */
	while (localflags) {
		onbit = localflags - (localflags & (localflags-1));
		localflags &= (localflags - 1);
		switch (onbit) {
		case TCP_STATS_SND_CWND:
			tcp_stats_buffer[tcp_stats_count++] = tp->snd_cwnd;
			break;
			
		case TCP_STATS_SND_MAX:
			tcp_stats_buffer[tcp_stats_count++] = tp->snd_max;
			break;
			
		case TCP_STATS_ISS:
			tcp_stats_buffer[tcp_stats_count++] = tp->iss;
			break;

		case TCP_STATS_IRS:
			tcp_stats_buffer[tcp_stats_count++] = tp->irs;
			break;
			
		case TCP_STATS_SND_SSTHRESH:
			tcp_stats_buffer[tcp_stats_count++] = tp->snd_ssthresh;
			break;
			
		case TCP_STATS_RXTCUR:
			tcp_stats_buffer[tcp_stats_count++] = tp->t_rxtcur;
			break;

		case TCP_STATS_RTT:
			tcp_stats_buffer[tcp_stats_count++] = tp->t_rtt;
			break;

		case TCP_STATS_SRTT:
			tcp_stats_buffer[tcp_stats_count++] = tp->t_srtt;
			break;
			
		case TCP_STATS_RTTVAR:
			tcp_stats_buffer[tcp_stats_count++] = tp->t_rttvar;
			break;

		case TCP_STATS_RCV_NXT:
			tcp_stats_buffer[tcp_stats_count++] = tp->rcv_nxt;
			break;

		case TCP_STATS_LAST_ACK_SENT:
			tcp_stats_buffer[tcp_stats_count++]=tp->last_ack_sent;
			break;

#ifdef ER
		case TCP_STATS_ER:
			tcp_stats_buffer[tcp_stats_count++] =
				tp->snd_nxt - tp->iss;
			break;
#endif

#ifdef FINE_GRAINED_TSTAMP
		case TCP_STATS_SRTT_EXACT:
			tcp_stats_buffer[tcp_stats_count++] = tp->t_srtt_exact;
			break;
#endif
			
		case TCP_STATS_COARSE_RXT:
		case TCP_STATS_DUPACK_RXT:
		case TCP_STATS_SMART_RXT:
		case TCP_STATS_ELN_RXT:  
		case TCP_STATS_DUPACK_EVENT:  
			break;

#ifdef ACC
		case TCP_STATS_DELACK:
			tcp_stats_buffer[tcp_stats_count++] = tp->delack;
			break;
#endif
			
		default:
			break;
		}
	}
}

/*
 * Dump the contents of tcp_stats_buffer to a file.
 */
void tcp_stats_dump() {
	/* XXXX */
}
