/*-
 * Copyright (c) 1996 Berkeley Software Design, Inc. All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *      BSDI ktr.h,v 1.3 1996/10/11 16:02:34 pjd Exp
 */

/*
 *	Wraparound kernel trace buffer.
 *
 *	To enable tracing globally use 'options KTR' in the config file.
 *	The buffer is sized at 256 entries, this can be overridden with
 *	'options KTR_SIZE'.
 *
 *	This tracing is intended to be lightweight enough to run in a
 *	production environment (if need be). It is used much like
 *	printf, however for each trace call (actually inline via a
 *	macro) only a trace entry (5 words) is updated (the time, a
 *	pointer to the printf style string, and two parameter words).
 *
 *  Usage:
 *    Unconditional traces:
 *	TR(fmt, p1, p2)		- Simple trace entry
 *	HTR(fmt, p2, p2)	- Trace entry with high-res timing
 *	BTR(fmt, p1, p2)	- Trace entry with cli/sti protection
 *				   (don't use in code that is already cli'd)
 *
 *    Conditional (entry made if (type & ktr_mask) != 0 at run time):
 *	CTR(type, fmt, p1, p2)	- Simple trace entry
 *	CBTR(type, fmt, p1, p2)	- Conditional entry with cli/sti protection
 *
 *  Notes:
 *	'fmt' is a pointer to a character string, usually generated by the
 *	compiler (as in "xxxx"). Only the pointer is stored, if the string
 *	is not constant then the result when the buffer is formatted may
 *	not be what is expected.
 *
 *	'p1' and 'p2' may be any small (4 bytes or less) data type, they
 *	are cast into an unsigned long.
 *
 *	The blocking form of the trace macros is used when there is conflict
 *	between top level and bottom level code (the window that is protected
 *	is the acquisition of a new trace entry).
 *
 *	High-res timing involves a call to microtime() and is somewhat slower
 *	than the other forms. This is most useful timing bcopy's or insw/outsw's
 *	to devices.
 *
 *	The trace classes are used to turn tracing on and off by manipulating
 *	ktr_mask. This facility can be used to enable/disable tracing to
 *	reduce clutter or to turn off all tracing the first time some condition
 *	occurs.
 *
 *	The trace buffer can be inspected by hand with gdb, by using the
 *	tdump script for gdb, or by running the tdump program (this is the
 *	preferred method).
 */

#ifndef _SYS_KTR_H_
#define _SYS_KTR_H_

/* Requires sys/types.h and sys/time.h */

#ifdef KERNEL
extern struct timeval time;
#endif

struct ktr_entry {
	struct	timeval ktr_tv;
	char	*ktr_desc;
	u_long	ktr_parm1;
	u_long	ktr_parm2;
};

#ifdef KTR
extern struct ktr_entry ktr_buf[];
extern volatile int ktr_idx;
extern int ktr_mask;
extern int ktr_size;

#define _TR(_desc, _p1, _p2) { \
	volatile struct ktr_entry *_ktrptr;			\
	int _idx;						\
	_idx = ktr_idx;						\
	_ktrptr = &ktr_buf[_idx];				\
	ktr_idx = (_idx + 1) & (ktr_size - 1);			\
	_ktrptr->ktr_tv.tv_sec = time.tv_sec;			\
	_ktrptr->ktr_tv.tv_usec = time.tv_usec;			\
	_ktrptr->ktr_desc = (_desc);				\
	_ktrptr->ktr_parm1 = (u_long)(_p1);			\
	_ktrptr->ktr_parm2 = (u_long)(_p2);			\
}

#define	TR(_desc, _p1, _p2) if (ktr_mask & KTR_GEN) _TR(_desc, _p1, _p2)
#define	CTR(_type, _desc, _p1, _p2) if (ktr_mask & _type) _TR(_desc, _p1, _p2)

#define _TRB(_desc, _p1, _p2) { \
	volatile struct ktr_entry *_ktrptr;			\
	int _idx;						\
	int _s = splmem_fast();					\
	_idx = ktr_idx;						\
	ktr_idx = (_idx + 1) & (ktr_size - 1);			\
	splxmem_fast(_s);					\
	_ktrptr = &ktr_buf[_idx];				\
	_ktrptr->ktr_tv.tv_sec = time.tv_sec;			\
	_ktrptr->ktr_tv.tv_usec = time.tv_usec;			\
	_ktrptr->ktr_desc = (_desc);				\
	_ktrptr->ktr_parm1 = (u_long)(_p1);			\
	_ktrptr->ktr_parm2 = (u_long)(_p2);			\
}

#define	BTR(_desc, _p1, _p2) if (ktr_mask & KTR_GEN) _TRB(_desc, _p1, _p2)
#define	CBTR(_type, _desc, _p1, _p2) if (ktr_mask & _type) _TRB(_desc, _p1, _p2)

#define HTR(_desc, _p1, _p2) { \
	volatile struct ktr_entry *_ktrptr;			\
	int _idx = ktr_idx;					\
	_ktrptr = &ktr_buf[_idx];				\
	ktr_idx = (_idx + 1) & (ktr_size - 1);			\
	microtime(&_ktrptr->ktr_tv);				\
	_ktrptr->ktr_desc = (_desc);				\
	_ktrptr->ktr_parm1 = (u_long)(_p1);			\
	_ktrptr->ktr_parm2 = (u_long)(_p2);			\
}
#else	/* KTR */
#define	TR(_desc, _p1, _p2)
#define	CTR(_type, _desc, _p1, _p2)
#define	BTR(_desc, _p1, _p2)
#define	CBTR(_type, _desc, _p1, _p2)
#define	HTR(_desc, _p1, _p2)
#endif	/* KTR */

/*
 * Trace classes
 */
#define	KTR_GEN		0x0001		/* General (TR) */
#define	KTR_INTR	0x0002		/* Interrupt tracing */
#define	KTR_IO		0x0004		/* Upper I/O  */
#define	KTR_FS		0x0008		/* Filesystem */
#define	KTR_DEV		0x0010		/* Device driver */
#define	KTR_PROC	0x0020		/* Process scheduling */
#define	KTR_SYSC	0x0040		/* System call */

#endif /* !_SYS_KTR_H_ */
